""""""
from __future__ import annotations

from flask import Flask

from .extension import AttachmentExtension, AttachmentsManager
from .forms import AttachmentForm


def register_plugin(app: Flask):
    AttachmentExtension(app)



"""Lightweight integration and denormalisation using events (signals)."""

from __future__ import annotations

from typing import Any

from blinker import ANY

from abilian.core.entities import Entity
from abilian.core.models.subjects import User
from abilian.core.signals import activity
from abilian.sbe.apps.documents.models import Document

from .models import Community


@activity.connect_via(ANY)
def update_community(
    sender: Any, verb: str, actor: User, object: Entity, target: Entity | None = None
) -> None:
    if isinstance(object, Community):
        object.touch()
        return

    if isinstance(target, Community):
        community = target
        community.touch()

        if isinstance(object, Document):
            if verb == "post":
                community.document_count += 1
            elif verb == "delete":
                community.document_count -= 1



from __future__ import annotations

import pytest
from pytest import fixture
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Session

from abilian.sbe.app import Application
from abilian.sbe.apps.documents.models import Document, Folder
from abilian.sbe.apps.documents.repository import ContentRepository


@fixture()
def repository(app: Application) -> ContentRepository:
    repository = ContentRepository()
    repository.init_app(app)
    return repository


@fixture()
def root(session: Session) -> Folder:
    root = Folder(title="")
    session.add(root)
    session.flush()
    return root


def test_create_doc(root: Folder, session: Session) -> None:
    doc = root.create_document("doc")
    session.flush()
    assert doc.path == "/doc"
    assert len(root.children) == 1
    assert not doc.is_root_folder


def test_create_folder(root: Folder, session: Session) -> None:
    folder = root.create_subfolder("folder")
    session.flush()
    assert folder.title == "folder"
    assert folder.name == "folder"
    assert folder.path == "/folder"
    assert len(root.children) == 1
    assert len(folder.children) == 0
    assert not folder.is_root_folder


def test_move(root: Folder, repository: ContentRepository) -> None:
    doc = root.create_document("doc")
    folder = root.create_subfolder("folder")

    repository.move_object(doc, folder, "newdoc")
    assert doc in folder.children
    assert doc not in root.children
    assert len(root.children) == 1
    assert len(folder.children) == 1
    assert doc.title == "newdoc"
    assert doc.path == "/folder/newdoc"


def test_copy(root: Folder, repository: ContentRepository) -> None:
    doc = root.create_document("doc")
    folder = root.create_subfolder("folder")

    doc_copy = repository.copy_object(doc, folder, "copydoc")

    assert len(root.children) == 2
    assert len(folder.children) == 1
    assert doc_copy.title == "copydoc"
    assert doc_copy.name == "copydoc"
    assert doc_copy.path == "/folder/copydoc"
    assert doc_copy in folder.children


def test_copy_nested_folders(root: Folder, repository: ContentRepository) -> None:
    folder1 = root.create_subfolder("folder1")
    folder2 = root.create_subfolder("folder2")
    subfolder = folder1.create_subfolder("subfolder")

    folder1_copy = repository.copy_object(folder1, folder2)

    assert len(root.children) == 2
    assert folder1 in root.children
    assert folder2 in root.children

    assert len(folder1.children) == 1
    assert subfolder in folder1.children

    assert len(folder2.children) == 1
    assert folder1_copy in folder2.children


def test_move_nested_folders(root: Folder, repository: ContentRepository) -> None:
    folder1 = root.create_subfolder("folder1")
    folder2 = root.create_subfolder("folder2")
    subfolder = folder1.create_subfolder("subfolder")

    repository.move_object(folder1, folder2)

    assert len(root.children) == 1
    assert len(folder1.children) == 1


def test_rename(root: Folder, repository: ContentRepository) -> None:
    folder = root.create_subfolder("folder")
    doc = folder.create_document("doc")

    repository.rename_object(doc, "doc1")
    assert doc.title == "doc1"
    assert doc.name == "doc1"
    assert doc.path == "/folder/doc1"

    repository.rename_object(folder, "folder1")
    assert folder.title == "folder1"
    assert folder.name == "folder1"
    assert folder.path == "/folder1"
    assert doc.path == "/folder1/doc1"


def test_delete(root: Folder, repository: ContentRepository, session: Session) -> None:
    doc = root.create_document("doc")
    folder = root.create_subfolder("folder")
    session.flush()

    assert doc.parent == root
    assert folder.parent == root
    assert doc in root.children
    assert folder in root.children
    assert not doc.is_root_folder
    assert not folder.is_root_folder

    doc_id = doc.id
    folder_id = folder.id

    repository.delete_object(doc)
    repository.delete_object(folder)
    assert len(root.children) == 0

    session.flush()
    assert Folder.query.get(folder_id) is None
    assert Document.query.get(doc_id) is None

    # test delete tree
    folder = root.create_subfolder("folder")
    sub = folder.create_subfolder("subfolder")
    doc = folder.create_document("doc")
    session.flush()

    repository.delete_object(folder)
    assert len(root.children) == 0
    assert Folder.query.all() == [root]
    assert Document.query.all() == []


def test_no_duplicate_name(root: Folder, session: Session) -> None:
    root.create_subfolder("folder_1")
    root.create_subfolder("folder_1")
    with pytest.raises(IntegrityError):
        session.flush()



from __future__ import annotations

import bleach
from wtforms import StringField, TextAreaField, ValidationError
from wtforms.fields.html5 import URLField

from abilian.i18n import _l
from abilian.web.forms import Form
from abilian.web.forms.fields import DateTimeField
from abilian.web.forms.filters import strip
from abilian.web.forms.validators import required
from abilian.web.forms.widgets import RichTextWidget

ALLOWED_TAGS = {
    "a",
    "abbr",
    "acronym",
    "b",
    "blockquote",
    "br",
    "code",
    "em",
    "i",
    "li",
    "ol",
    "strong",
    "ul",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "p",
    "u",
    "img",
}

ALLOWED_ATTRIBUTES = {
    "*": ["title"],
    "p": ["style"],
    "a": ["href", "title"],
    "abbr": ["title"],
    "acronym": ["title"],
    "img": ["src", "alt", "title"],
}

ALLOWED_STYLES = ["text-align"]

WIDGET_ALLOWED = {}
for attr in ALLOWED_TAGS:
    allowed = ALLOWED_ATTRIBUTES.get(attr, True)
    if not isinstance(allowed, bool):
        allowed = {tag: True for tag in allowed}
    WIDGET_ALLOWED[attr] = allowed


class EventForm(Form):
    title = StringField(label=_l("Title"), filters=(strip,), validators=[required()])

    start = DateTimeField(_l("Start"), validators=[required()])
    end = DateTimeField(_l("End"), validators=[required()])

    location = TextAreaField(label=_l("Location"), filters=(strip,))

    url = URLField(label=_l("URL"), filters=(strip,))

    description = TextAreaField(
        label=_l("Description"),
        widget=RichTextWidget(allowed_tags=WIDGET_ALLOWED),
        filters=(strip,),
        validators=[required()],
    )

    def validate_description(self, field):
        field.data = bleach.clean(
            field.data,
            tags=ALLOWED_TAGS,
            attributes=ALLOWED_ATTRIBUTES,
            styles=ALLOWED_STYLES,
            strip=True,
        )

    def validate_end(self, field):
        if self.start.data > self.end.data:
            raise ValidationError(_l("End date/time must be after start"))


EventForm.start.kwargs["raw_data"] = [" | 09:00"]
EventForm.end.kwargs["raw_data"] = [" | 18:00"]



"""Create an application instance.

(wsgi script inside the package extranet, expecting no change in sys.path)
"""

# flake8: noqa

# Temps monkey patches
import werkzeug.datastructures
import werkzeug.urls

werkzeug.url_encode = werkzeug.urls.url_encode
werkzeug.FileStorage = werkzeug.datastructures.FileStorage

# Normal bootstrap
from flask.cli import load_dotenv

from .app import create_app

load_dotenv()

app = create_app()



"""Dramatiq tasks related to documents."""
from __future__ import annotations

from collections.abc import Iterator
from contextlib import contextmanager
from typing import TYPE_CHECKING

from loguru import logger
from sqlalchemy.orm import Session

from abilian.core.dramatiq.singleton import dramatiq
from abilian.core.extensions import db
from abilian.services import converter, get_service
from abilian.services.conversion import ConversionError, HandlerNotFound

if TYPE_CHECKING:
    from .models import Document


@contextmanager
def get_document(
    document_id: int, session: Session | None = None
) -> Iterator[tuple[Session, Document | None]]:
    """Context manager that yields (session, document)."""
    from .models import Document

    #
    logger.info(f"get_document() {document_id=} {session=}")

    if session is None:
        doc_session = db.create_scoped_session()
    else:
        doc_session = session

    with doc_session.begin_nested():
        query = doc_session.query(Document)
        document = query.get(document_id)
        logger.debug(f"get_document() {document=} {session=}")
        yield (doc_session, document)

    # cleanup
    if session is None:
        doc_session.commit()
        doc_session.close()


@dramatiq.actor(max_retries=20, max_backoff=86400000)
def process_document(document_id: int) -> None:
    """Task to process a document.

    max_retries = 20 (Dramatiq default)
    max_backoff = 86400000 , i.e. 1 day
    """
    logger.info(f"process_document() actor : {document_id=}")

    with get_document(document_id) as (session, document):
        if document is None:
            return

        # True = Ok, None means no check performed (no antivirus present)
        is_clean = _run_antivirus(document)
        logger.debug(f"process_document() {document_id=} {is_clean=}")
        if is_clean is False:
            return
    logger.debug(f"process_document() virus clean {document_id=}")
    preview_document.send(document_id)
    logger.debug(f"process_document() before convert_document_content {document_id=}")
    convert_document_content.send(document_id)
    logger.debug(f"process_document() exit {document_id=}")


def _run_antivirus(document: Document) -> bool | None:
    antivirus = get_service("antivirus")
    if antivirus and antivirus.running:
        is_clean = antivirus.scan(document.content_blob)
        if "antivirus_task" in document.content_blob.meta:
            del document.content_blob.meta["antivirus_task"]
        return is_clean
    return None


@dramatiq.actor
def antivirus_scan(document_id):
    """Return antivirus.scan() result."""
    with get_document(document_id) as (session, document):
        if document is None:
            return None
        return _run_antivirus(document)


@dramatiq.actor(max_retries=5)
def preview_document(document_id: int) -> None:
    """Compute the document preview images with its default preview size."""

    logger.info(f"preview_document() {document_id=}")

    with get_document(document_id) as (session, document):
        logger.debug(f"preview_document() {document_id=} {document=}")
        if document is None:
            # deleted after task queued, but before task run
            return

        convert_to_image(document)


@logger.catch(level="ERROR")
def convert_to_image(doc: Document) -> None:
    logger.info(f"convert_to_image() {doc=}")

    try:
        converter.to_image(
            doc.content_digest,
            doc.content,
            doc.content_type,
            0,
            doc.preview_size,
        )
    except ConversionError as e:
        logger.info(f"Preview failed for document {doc.name}: {e}")


@dramatiq.actor(max_retries=5)
def convert_document_content(document_id: int) -> None:
    """Convert document content."""

    logger.info(f"convert_document_content() {document_id=}")

    with get_document(document_id) as (session, doc):
        if doc is None:
            # deleted after task queued, but before task run
            return

        convert_to_pdf(doc)
        convert_to_text(doc)
        extract_metadata(doc)


@logger.catch(level="ERROR")
def convert_to_pdf(doc: Document) -> None:
    logger.info(f"convert_to_pdf() {doc=}")

    if doc.content_type == "application/pdf":
        doc.pdf = doc.content
        return
    try:
        doc.pdf = converter.to_pdf(doc.content_digest, doc.content, doc.content_type)
    except (HandlerNotFound, ConversionError) as e:
        doc.pdf = b""
        logger.info(f"Conversion to PDF failed for document {doc.name}: {e}")


@logger.catch(level="ERROR")
def convert_to_text(doc: Document) -> None:
    logger.info(f"convert_to_text() {doc=}")

    try:
        doc.text = converter.to_text(doc.content_digest, doc.content, doc.content_type)
    except ConversionError as e:
        doc.text = ""
        logger.info(f"Conversion to text failed for document {doc.name}: {e}")


@logger.catch(level="ERROR")
def extract_metadata(doc: Document) -> None:
    logger.info(f"extract_metadata() {doc=}")

    doc.extra_metadata = {}
    try:
        doc.extra_metadata = converter.get_metadata(
            doc.content_digest, doc.content, doc.content_type
        )
    except ConversionError as e:
        logger.warning(f"Metadata extraction failed on document {doc.name}: {e}")
    except UnicodeDecodeError as e:
        logger.error(f"Unicode issue on document {doc.name}: {e}")
    except Exception as e:
        logger.error(f"Other issue on document {doc.name}: {e}")

    if doc.text:
        import langid

        doc.language = langid.classify(doc.text)[0]

    doc.page_num = doc.extra_metadata.get("PDF:Pages", 1)



"""Create all standard extensions."""

# Note: Because of issues with circular dependencies, Abilian-specific
# extensions are created later.
from __future__ import annotations

import sqlite3
from typing import Any

import flask_mail
import sqlalchemy as sa
import sqlalchemy.event
import sqlalchemy.orm
from flask import current_app
from loguru import logger
from sqlalchemy.engine import Connection, Engine
from sqlalchemy.sql.schema import MetaData

from abilian.core.sqlalchemy import SQLAlchemy

from . import upstream_info
from .csrf import abilian_csrf
from .csrf import wtf_csrf as csrf
from .jinjaext import DeferredJS
from .login import login_manager
from .redis import Redis

__all__ = (
    "csrf",
    "db",
    "get_extension",
    "login_manager",
    "mail",
    "redis",
    "upstream_info",
)


# patch flask.ext.mail.Message.send to always set enveloppe_from default mail
# sender
# FIXME: we'ld rather subclass Message and update all imports
def _message_send(self: Any, connection: flask_mail.Connection):
    """Send a single message instance.

    If TESTING is True the message will not actually be sent.

    :param self: a Message instance.
    """
    sender = current_app.config["MAIL_SENDER"]
    if not self.extra_headers:
        self.extra_headers = {}
    self.extra_headers["Sender"] = sender
    connection.send(self, sender)


logger.info(flask_mail.Message.send)
flask_mail.Message.send = _message_send

mail = flask_mail.Mail()

db = SQLAlchemy()

redis = Redis()

deferred_js = DeferredJS()


@sa.event.listens_for(db.metadata, "before_create")
@sa.event.listens_for(db.metadata, "before_drop")
def _filter_metadata_for_connection(
    target: MetaData, connection: Connection, **kw: Any
):
    """Listener to control what indexes get created.

    Useful for skipping postgres-specific indexes on a sqlite for example.

    It's looking for info entry `engines` on an index
    (`Index(info=dict(engines=['postgresql']))`), an iterable of engine names.
    """
    engine = connection.engine.name
    default_engines = (engine,)
    tables = target if isinstance(target, sa.Table) else kw.get("tables", [])
    for table in tables:
        indexes = list(table.indexes)
        for idx in indexes:
            if engine not in idx.info.get("engines", default_engines):
                table.indexes.remove(idx)


#
# CSRF
#
def get_extension(name: str):
    """Get the named extension from the current app, returning None if not
    found."""
    return current_app.extensions.get(name)


def _install_get_display_value(cls: Any):
    _MARK = object()

    def display_value(self, field_name, value=_MARK):
        """Return display value for fields having 'choices' mapping (stored
        value.

        -> human readable value). For other fields it will simply return field
        value.

        `display_value` should be used instead of directly getting field value.

        If `value` is provided, it is "translated" to a human-readable value. This is
        useful for obtaining a human readable label from a raw value.
        """
        val = getattr(self, field_name, "ERROR") if value is _MARK else value

        mapper = sa.orm.object_mapper(self)
        try:
            field = getattr(mapper.c, field_name)
        except AttributeError:
            pass
        else:
            if "choices" in field.info:

                def get(v):
                    return field.info["choices"].get(v, v)

                if isinstance(val, list):
                    val = [get(v) for v in val]
                else:
                    val = get(val)

        return val

    if not hasattr(cls, "display_value"):
        cls.display_value = display_value


sa.event.listen(db.Model, "class_instrument", _install_get_display_value)


#
# Make Sqlite a bit more well-behaved.
#
@sa.event.listens_for(Engine, "connect")
def _set_sqlite_pragma(dbapi_connection: Any, connection_record: Any):
    if isinstance(dbapi_connection, sqlite3.Connection):  # pragma: no cover
        cursor = dbapi_connection.cursor()
        cursor.execute("PRAGMA foreign_keys=ON;")
        cursor.close()



"""Not used!"""

from __future__ import annotations

from flask import flash, redirect, render_template, request, url_for
from flask_babel import gettext as _
from flask_login import current_user
from flask_mail import Message as Email
from sqlalchemy.schema import Column
from sqlalchemy.types import Text

from abilian.core.extensions import db, mail

from .social import social


class Invite(db.Model):
    """Tracks a pending invitation to join the community."""

    __tablename__ = "invitation"

    email = Column(Text, nullable=False)

    query = db.session.query_property()


@social.route("/invite")
def invite():
    return render_template("social/invite.html")


@social.route("/invite", methods=["POST"])
def invite_post():
    action = request.form.get("action", "cancel")
    if action == "cancel":
        flash(_("Action aborted"), "info")
        return redirect(url_for(".home"))

    emails = request.form.get("emails").split("\n")
    emails = [email.strip() for email in emails]

    # FIXME: what do we do with this ?
    message = request.form.get("message", "")

    with mail.connect() as conn:
        for email in emails:
            # FIXME: what do we do with this ?
            invite = Invite(sender=current_user, email=email)

            # subject = _(u"%s would like to invite you to the %s community")
            # % (current_user.name, "Yaka")
            subject = "{} would like to invite you to the {} community".format(
                current_user.name, "Yaka"
            )
            msg = Email(subject, recipients=[email], sender=current_user.email)
            params = {"org_name": "Yaka"}
            msg.body = render_template("social/mail/invite.txt", **params)
            conn.send(msg)

    flash(_("Invitation sent"), "info")
    return redirect(url_for(".home"))



from __future__ import annotations

from urllib.parse import urlparse

from flask import url_for
from flask.testing import FlaskClient

from abilian.core.sqlalchemy import SQLAlchemy
from abilian.sbe.app import Application
from abilian.sbe.apps.communities.models import Community
from abilian.services import security_service
from abilian.services.security import Admin
from abilian.testing.util import client_login


def test_index(
    community1: Community, app: Application, db: SQLAlchemy, client: FlaskClient
) -> None:
    security_service.start()

    user = community1.test_user
    with client_login(client, user):
        response = client.get(url_for("communities.index"))
        assert response.status_code == 200


def test_community_home(
    community1: Community, community2: Community, app: Application, client: FlaskClient
) -> None:
    security_service.start()

    url = app.default_view.url_for(community1)

    user1 = community1.test_user

    with client_login(client, user1):
        response = client.get(url)
        assert response.status_code == 302
        expected_url = url_for(
            "wall.index", community_id=community1.slug, _external=True
        )
        assert urlparse(response.location).path == urlparse(expected_url).path
        # assert response.location == expected_url

    user2 = community2.test_user
    with client_login(client, user2):
        response = client.get(url)
        assert response.status_code == 403


def test_new(
    community1: Community, app: Application, client: FlaskClient, db: SQLAlchemy
) -> None:
    # security_service.use_cache = False
    security_service.start()

    user = community1.test_user

    with client_login(client, user):
        response = client.get(url_for("communities.new"))
        assert response.status_code == 403

    security_service.grant_role(user, Admin)
    db.session.flush()

    with client_login(client, user):
        response = client.get(url_for("communities.new"))
        assert response.status_code == 200


def test_community_settings(
    app: Application, client: FlaskClient, community1: Community
) -> None:
    security_service.start()

    url = url_for("communities.settings", community_id=community1.slug)
    user = community1.test_user

    with client_login(client, user):
        response = client.get(url)
        assert response.status_code == 403

        security_service.grant_role(user, Admin)
        response = client.get(url)
        assert response.status_code == 200

        data = {
            "__action": "edit",
            "name": "edited community",
            "description": "my community",
            "linked_group": "",
            "type": "participative",
        }
        response = client.post(url, data=data, follow_redirects=True)
        assert response.status_code == 200
        assert "edited community" in response.get_data(as_text=True)


def test_members(
    app: Application,
    client: FlaskClient,
    db: SQLAlchemy,
    community1: Community,
    community2: Community,
) -> None:
    security_service.start()

    user1 = community1.test_user
    user2 = community2.test_user

    with client_login(client, user1):
        url = url_for("communities.members", community_id=community1.slug)
        response = client.get(url)
        assert response.status_code == 200

        # test add user
        data = {"action": "add-user-role", "user": user2.id}
        response = client.post(url, data=data)
        assert response.status_code == 403

        security_service.grant_role(user1, Admin)

        data = {"action": "add-user-role", "user": user2.id, "role": "member"}
        response = client.post(url, data=data, follow_redirects=True)
        assert response.status_code == 200

        membership = next(m for m in community1.memberships if m.user == user2)
        assert membership.role == "member"

        data["action"] = "set-user-role"
        data["role"] = "manager"
        response = client.post(url, data=data, follow_redirects=True)
        assert response.status_code == 200

        db.session.expire(membership)
        assert membership.role == "manager"

        # Community.query.session is not self.db.session, but web app
        # session.
        community = Community.query.get(community1.id)
        assert user2 in community.members

        # test delete
        data = {
            "action": "delete",
            "user": user2.id,
            "membership": next(m.id for m in community1.memberships if m.user == user2),
        }
        response = client.post(url, data=data, follow_redirects=True)
        assert response.status_code == 200
        assert user2 not in community.members



from __future__ import annotations

from datetime import datetime, timedelta

import html2text
from flask import current_app
from flask_mail import Message
from loguru import logger
from sqlalchemy import and_, or_
from validate_email import validate_email

from abilian.core.dramatiq.scheduler import crontab
from abilian.core.dramatiq.singleton import dramatiq
from abilian.core.entities import Entity
from abilian.core.models.subjects import User
from abilian.core.util import md5
from abilian.i18n import render_template_i18n
from abilian.sbe.apps.communities.models import Community
from abilian.sbe.apps.documents.models import Document
from abilian.sbe.apps.documents.repository import content_repository
from abilian.sbe.apps.forum.models import Post, Thread
from abilian.sbe.apps.wiki.models import WikiPage
from abilian.services import get_service
from abilian.services.activity import ActivityEntry
from abilian.services.auth.views import get_serializer
from abilian.web import url_for

from .. import TOKEN_SERIALIZER_NAME


@crontab("SCHEDULE_SEND_DAILY_SOCIAL_DIGEST")
@dramatiq.actor()
def send_daily_social_digest_task():
    # a request_context is required when rendering templates

    logger.info("Running job: send_daily_social_digest_task")
    with current_app.test_request_context("/tasks/send_daily_social_updates"):
        config = current_app.config
        if not config.get("PRODUCTION") or config.get("DEMO"):
            return

        send_daily_social_digest()


def send_daily_social_digest():
    for user in User.query.filter(User.can_login == True).all():
        preferences = get_service("preferences")
        prefs = preferences.get_preferences(user)

        if not prefs.get("sbe:notifications:daily", False):
            continue

        # Defensive programming.
        if not validate_email(user.email):
            continue

        try:
            send_daily_social_digest_to(user)
        except BaseException:
            current_app.logger.error("Error sending daily social digest", exc_info=True)


def send_daily_social_digest_to(user: User):
    """Send to a given user a daily digest of activities in its communities.

    Return 1 if mail sent, 0 otherwise.
    """
    mail = current_app.extensions["mail"]

    message = make_message(user)
    if message:
        mail.send(message)
        return 1
    else:
        return 0


def make_message(user: User) -> Message | None:
    config = current_app.config
    sender = config.get("BULK_MAIL_SENDER", config["MAIL_SENDER"])
    sbe_config = config["ABILIAN_SBE"]
    subject = sbe_config["DAILY_SOCIAL_DIGEST_SUBJECT"]

    recipient = user.email
    digests = []
    happened_after = datetime.utcnow() - timedelta(days=1)
    list_id = '"{} daily digest" <daily.digest.{}>'.format(
        config["SITE_NAME"], config.get("SERVER_NAME", "example.com")
    )
    base_extra_headers = {
        "List-Id": list_id,
        "List-Post": "NO",
        "Auto-Submitted": "auto-generated",
        "X-Auto-Response-Suppress": "All",
        "Precedence": "bulk",
    }

    for membership in user.communautes_membership:
        community = membership.community
        if not community:
            # TODO: should not happen but it does. Fix root cause instead.
            continue
        # create an empty digest
        digest = CommunityDigest(community)
        AE = ActivityEntry
        activities = (
            AE.query.order_by(AE.happened_at.asc())
            .filter(
                and_(
                    AE.happened_at > happened_after,
                    or_(
                        and_(
                            AE.target_type == community.object_type,
                            AE.target_id == community.id,
                        ),
                        and_(
                            AE.object_type == community.object_type,
                            AE.object_id == community.id,
                        ),
                    ),
                )
            )
            .all()
        )

        # fill the internal digest lists with infos
        # seen_entities, new_members, new_documents, updated_documents ...
        for activity in activities:
            digest.update_from_activity(activity, user)
        # if activities:
        #   import ipdb; ipdb.set_trace()
        # save the current digest in the master digests list
        if not digest.is_empty():
            digests.append(digest)

    if not digests:
        return None

    token = generate_unsubscribe_token(user)
    unsubscribe_url = url_for(
        "notifications.unsubscribe_sbe",
        token=token,
        _external=True,
        _scheme=config["PREFERRED_URL_SCHEME"],
    )
    extra_headers = dict(base_extra_headers)
    extra_headers["List-Unsubscribe"] = f"<{unsubscribe_url}>"

    msg = Message(
        subject, sender=sender, recipients=[recipient], extra_headers=extra_headers
    )
    ctx = {"digests": digests, "token": token, "unsubscribe_url": unsubscribe_url}
    msg.html = render_template_i18n("notifications/daily-social-digest.html", **ctx)
    msg.body = html2text.html2text(msg.html)
    return msg


def generate_unsubscribe_token(user: User) -> str:
    """Generates a unique unsubscription token for the specified user.

    :param user: The user to work with
    """
    data = [str(user.id), md5(user.password)]
    return get_serializer(TOKEN_SERIALIZER_NAME).dumps(data)


class CommunityDigest:
    def __init__(self, community: Community) -> None:
        self.community = community

        self.seen_entities: set[Entity] = set()
        self.new_members: list[User] = []
        self.new_documents: list[Document] = []
        self.updated_documents: list[Document] = []
        self.new_conversations: list[Post] = []
        self.updated_conversations: dict[Post, dict] = {}
        self.new_wiki_pages: list[WikiPage] = []
        self.updated_wiki_pages: dict[WikiPage, dict] = {}

    def is_empty(self):
        return (
            not self.new_members
            and not self.new_documents
            and not self.updated_documents
            and not self.new_conversations
            and not self.updated_conversations
            and not self.new_wiki_pages
            and not self.updated_wiki_pages
        )

    def update_from_activity(self, activity, user):
        actor = activity.actor
        obj = activity.object

        # TODO ?
        # target = activity.target

        if activity.verb == "join":
            self._update_for_join(actor)

        elif activity.verb == "post":
            self._update_for_post(actor, obj, user)

        elif activity.verb == "update":
            self._update_for_update(actor, obj, user)

    def _update_for_join(self, actor):
        self.new_members.append(actor)

    def _update_for_post(self, actor, obj, user):
        if obj is None:
            return
        if obj.id in self.seen_entities:
            return

        self.seen_entities.add(obj.id)

        if isinstance(obj, Document) and content_repository.has_access(user, obj):
            self.new_documents.append(obj)
        elif isinstance(obj, WikiPage):
            self.new_wiki_pages.append(obj)
        elif isinstance(obj, Thread):
            self.new_conversations.append(obj)
        elif isinstance(obj, Post):
            if obj.thread.id not in self.seen_entities:
                # save actor and oldest/first modified Post in thread
                # oldest post because Activities are ordered_by
                # Asc(A.happened_at)
                self.updated_conversations[obj.thread] = {
                    "actors": [actor],
                    "post": obj,
                }
                # Mark this post's Thread as seen to avoid duplicates
                self.seen_entities.add(obj.thread.id)
            elif obj.thread not in self.new_conversations:
                # this post's Thread has already been seen in another Activity
                # exclude it to avoid duplicates but save the Post's actor
                self.updated_conversations[obj.thread]["actors"].append(actor)

    def _update_for_update(self, actor, obj, user):
        if obj is None:
            return
        # special case for Wikipage, we want to know each updater
        if isinstance(obj, WikiPage):
            if obj in self.updated_wiki_pages:
                page = self.updated_wiki_pages[obj]
                if actor in page:
                    page[actor] += 1
                else:
                    page[actor] = 1
            else:
                self.updated_wiki_pages[obj] = {actor: 1}

        # fast return for all other objects
        if obj.id in self.seen_entities:
            return
        self.seen_entities.add(obj.id)

        # all objects here need to be accounted only once
        if isinstance(obj, Document) and content_repository.has_access(user, obj):
            self.updated_documents.append(obj)



from __future__ import annotations

from collections.abc import Callable
from functools import wraps
from typing import TYPE_CHECKING, Any

from flask import current_app
from loguru import logger

from abilian.core.util import fqcn

if TYPE_CHECKING:
    from abilian.app import Application


class ServiceNotRegistered(Exception):
    pass


class ServiceState:
    """Service state stored in Application.extensions."""

    #: reference to :class:`Service` instance
    service: Service

    running = False

    def __init__(self, service: Service, running: bool = False):
        self.service = service
        self.running = running
        self.logger = logger


class Service:
    """Base class for services."""

    #: State class to use for this Service
    AppStateClass = ServiceState

    #: service name in Application.extensions / Application.services
    name = ""

    def __init__(self, app: Any | None = None):
        if self.name is None:
            msg = f"Service must have a name ({fqcn(self.__class__)})"
            raise ValueError(msg)

        self.logger = logger
        if app:
            self.init_app(app)

    def init_app(self, app: Application):
        app.extensions[self.name] = self.AppStateClass(self)
        app.services[self.name] = self

    def start(self, ignore_state: bool = False):
        """Starts the service."""
        self.logger.debug("Start service")
        self._toggle_running(True, ignore_state)

    def stop(self, ignore_state: bool = False):
        """Stops the service."""
        self.logger.debug("Stop service")
        self._toggle_running(False, ignore_state)

    def _toggle_running(self, run_state: bool, ignore_state: bool = False):
        state = self.app_state
        run_state = bool(run_state)
        if not ignore_state:
            assert run_state ^ state.running
        state.running = run_state

    @property
    def app_state(self) -> Any:
        """Current service state in current application.

        :raise:RuntimeError if working outside application context.
        """
        try:
            return current_app.extensions[self.name]
        except KeyError as e:
            raise ServiceNotRegistered(self.name) from e

    @property
    def running(self) -> bool:
        """
        :returns: `False` if working outside application context, if service is
            not registered on current application, or if service is halted
            for current application.
        """
        try:
            return self.app_state.running
        except (RuntimeError, ServiceNotRegistered):
            # RuntimeError: happens when current_app is None: working outside
            # application context
            return False

    @staticmethod
    def if_running(meth: Callable) -> Callable:
        """Decorator for service methods that must be ran only if service is in
        running state."""

        @wraps(meth)
        def check_running(self: Any, *args: Any, **kwargs: Any) -> Any | None:
            if not self.running:
                return None
            return meth(self, *args, **kwargs)

        return check_running



from __future__ import annotations

import os
from pathlib import Path
from typing import Any

from flask import Flask
from flask_assets import Bundle
from flask_assets import Environment as AssetsEnv
from loguru import logger

from abilian.services.security import Anonymous
from abilian.web.assets.filters import ClosureJS


class AssetManagerMixin(Flask):
    def init_assets(self):
        if self.debug:
            js_filters = None
        else:
            if os.system("java -version 2> /dev/null") == 0:
                js_filters = ("closure_js",)
            else:
                logger.warning("Java is not installed. Can't use Closure")
                js_filters = None

        self._assets_bundles = {
            "css": {
                "options": {
                    "filters": ("less", "cssmin"),
                    "output": "style-%(version)s.min.css",
                }
            },
            "js-top": {
                "options": {"output": "top-%(version)s.min.js", "filters": js_filters}
            },
            "js": {
                "options": {"output": "app-%(version)s.min.js", "filters": js_filters}
            },
        }

        # bundles for JS translations
        languages = self.config["BABEL_ACCEPT_LANGUAGES"]
        for lang in languages:
            code = f"js-i18n-{lang}"
            filename = f"lang-{lang}-%(version)s.min.js"
            self._assets_bundles[code] = {
                "options": {"output": filename, "filters": js_filters}
            }

    def setup_asset_extension(self):
        assets = self.extensions["webassets"] = AssetsEnv(self)
        if self.debug:
            assets.debug = True
        assets.requirejs_config = {"waitSeconds": 90, "shim": {}, "paths": {}}

        assets_base_dir = Path(self.instance_path, "webassets")
        assets_dir = assets_base_dir / "compiled"
        assets_cache_dir = assets_base_dir / "cache"
        for path in (assets_base_dir, assets_dir, assets_cache_dir):
            if not path.exists():
                path.mkdir()

        assets.directory = str(assets_dir)
        assets.cache = str(assets_cache_dir)
        manifest_file = assets_base_dir / "manifest.json"
        assets.manifest = f"json:{manifest_file}"

        # set up load_path for application static dir. This is required
        # since we are setting Environment.load_path for other assets
        # (like core_bundle below),
        # in this case Flask-Assets uses webasssets resolvers instead of
        # Flask's one
        assets.append_path(self.static_folder, self.static_url_path)

        # filters options
        less_args = ["-ru"]
        assets.config["less_extra_args"] = less_args
        assets.config["less_as_output"] = True
        if assets.debug:
            assets.config["less_source_map_file"] = "style.map"

        # setup static url for our assets
        from abilian.web import assets as core_bundles

        core_bundles.init_app(self)

        # static minified are here
        assets.url = f"{self.static_url_path}/min"
        assets.append_path(str(assets_dir), assets.url)
        self.add_static_url(
            "min", str(assets_dir), endpoint="webassets_static", roles=Anonymous
        )

    def _finalize_assets_setup(self):
        assets = self.extensions["webassets"]
        assets_dir = Path(assets.directory)
        closure_base_args = [
            "--jscomp_warning",
            "internetExplorerChecks",
            "--source_map_format",
            "V3",
            "--create_source_map",
        ]

        for name, data in self._assets_bundles.items():
            bundles = data.get("bundles", [])
            options: dict[str, Any] = data.get("options", {})
            filters = options.get("filters") or []
            options["filters"] = []
            for f in filters:
                if f == "closure_js":
                    js_map_file = str(assets_dir / f"{name}.map")
                    f = ClosureJS(extra_args=closure_base_args + [js_map_file])
                options["filters"].append(f)

            if not options["filters"]:
                options["filters"] = None

            if bundles:
                assets.register(name, Bundle(*bundles, **options))

    def register_asset(self, type_: str, *assets: Any):
        """Register webassets bundle to be served on all pages.

        :param type_: `"css"`, `"js-top"` or `"js""`.

        :param assets:
            a path to file, a :ref:`webassets.Bundle <webassets:bundles>`
            instance or a callable that returns a
            :ref:`webassets.Bundle <webassets:bundles>` instance.

        :raises KeyError: if `type_` is not supported.
        """
        supported = list(self._assets_bundles.keys())
        if type_ not in supported:
            msg = "Invalid type: {}. Valid types: {}".format(
                repr(type_), ", ".join(sorted(supported))
            )
            raise KeyError(msg)

        for asset in assets:
            if not isinstance(asset, Bundle) and callable(asset):
                asset = asset()

            self._assets_bundles[type_].setdefault("bundles", []).append(asset)

    def register_i18n_js(self, *paths: str):
        """Register templates path translations files, like
        `select2/select2_locale_{lang}.js`.

        Only existing files are registered.
        """
        languages = self.config["BABEL_ACCEPT_LANGUAGES"]
        assets = self.extensions["webassets"]

        for path in paths:
            for lang in languages:
                filename = path.format(lang=lang)
                try:
                    assets.resolver.search_for_source(assets, filename)
                except OSError:
                    pass
                    # logger.debug('i18n JS not found, skipped: "%s"', filename)
                else:
                    self.register_asset(f"js-i18n-{lang}", filename)

    def register_base_assets(self):
        """Register assets needed by Abilian.

        This is done in a separate method in order to allow applications
        to redefine it at will.
        """
        from abilian.web import assets as bundles

        self.register_asset("css", bundles.LESS)
        self.register_asset("js-top", bundles.TOP_JS)
        self.register_asset("js", bundles.JS)
        self.register_i18n_js(*bundles.JS_I18N)



"""Dramatiq tasks related to document transformation and preview."""
from __future__ import annotations

import email
import mailbox
import re
from os.path import expanduser
from pathlib import Path
from typing import Any

import bleach
import chardet
import html2text
from bleach.css_sanitizer import CSSSanitizer
from flask import current_app, g
from flask_babel import get_locale
from flask_mail import Message
from itsdangerous import Serializer
from loguru import logger

from abilian.core.dramatiq.scheduler import crontab
from abilian.core.dramatiq.setup import RATE_LIMITER
from abilian.core.dramatiq.singleton import dramatiq
from abilian.core.extensions import db, mail
from abilian.core.models.subjects import User
from abilian.core.signals import activity
from abilian.core.util import md5, unwrap
from abilian.i18n import _l, render_template_i18n
from abilian.web import url_for

from .forms import ALLOWED_ATTRIBUTES, ALLOWED_STYLES, ALLOWED_TAGS
from .models import Post, PostAttachment, Thread

MAIL_REPLY_MARKER = _l("_____Write above this line to post_____")


@dramatiq.actor(max_retries=20, max_backoff=86400000)
def send_post_by_email(post_id: int | str):
    """Send a post to community members by email.

    max_retries = 20 (Dramatiq default)
    max_backoff = 86400000 , i.e. 1 day
    """
    with current_app.test_request_context("/tasks/send_post_by_email"):
        post = Post.query.get(post_id)
        if post is None:
            # deleted after task queued, but before task run
            return

        thread = post.thread
        community = thread.community
        logger.info(
            f"Sending new post by email to members of community {community.name!r}"
        )

        CHUNK_SIZE = 20
        members_id = [member.id for member in community.members if member.can_login]
        chunk = []
        for idx, member_id in enumerate(members_id):
            chunk.append(member_id)
            if idx % CHUNK_SIZE == 0:
                batch_send_post_to_users(post.id, chunk)
                chunk = []

        if chunk:
            batch_send_post_to_users(post.id, chunk)


# FIXME: add the dramatiq backernd for rate limit
# @shared_task(max_retries=10, rate_limit="12/m")
@dramatiq.actor(max_retries=20, max_backoff=86400000)
def batch_send_post_to_users(post_id, members_id, failed_ids=None):
    """Task run from send_post_by_email; auto-retry for mails that could not be
    successfully sent.

    Task default rate limit is 6/min.: there is at least 5 seconds between 2
    batches.

    During retry, if all `members_id` fails again, the task is retried
    5min. later, then 10, 20, 40... up to 10 times before giving up. This ensures
    retries up to approximatively 3 days and 13 hours after initial attempt
    (geometric series is: 5min * (1-2**10) / 1-2) = 5115 mins).
    """
    if not members_id:
        return None

    post = Post.query.get(post_id)
    if post is None:
        # deleted after task queued, but before task run
        return None
    with RATE_LIMITER[0].acquire():
        logger.info(f"{post_id=} {members_id=} {failed_ids=}")
        failed = set()
        successfully_sent = []
        thread = post.thread
        community = thread.community
        user_filter = (
            User.id.in_(members_id) if len(members_id) > 1 else User.id == members_id[0]
        )
        users = User.query.filter(user_filter).all()

        for user in users:
            try:
                with current_app.test_request_context("/send_post_by_email"):
                    send_post_to_user(community, post, user)
            except BaseException:
                failed.add(user.id)
            else:
                successfully_sent.append(user.id)

        # if failed:
        #     if failed_ids is not None:
        #         failed_ids = set(failed_ids)
        #
        #     if failed == failed_ids:
        #         # 5 minutes * (2** retry count)
        #         countdown = 300 * 2 ** batch_send_post_to_users.request.retries
        #         batch_send_post_to_users.retry([post_id, list(failed)], countdown=countdown)
        #     else:
        #         batch_send_post_to_users.apply_async([post_id, list(failed)])

        return {
            "post_id": post_id,
            "successfully_sent": successfully_sent,
            "failed": list(failed),
        }


def build_local_part(name, uid):
    """Build local part as 'name-uid-digest', ensuring length < 64."""
    key = current_app.config["SECRET_KEY"]
    serializer = Serializer(key)
    signature = serializer.dumps(uid)
    digest = md5(signature)
    local_part = f"{name}+{uid}-{digest}"

    if len(local_part) > 64:
        if (len(local_part) - len(digest) - 1) > 64:
            # even without digest, it's too long
            raise ValueError(
                "Cannot build reply address: local part exceeds 64 characters"
            )
        local_part = local_part[:64]

    return local_part


def build_reply_email_address(name: str, post: Post, member: User, domain: str) -> str:
    """Build a reply-to email address embedding the locale, thread_id and
    user.id.

    :param name: (str)    first part of an email address
    :param post: Post()   to get post.thread_id
    :param member: User() to get user.id
    :param domain: (str)  the last domain name of the email address
    :return: (Unicode)    reply address for forum in the form
       test+P-fr-3-4-SDB7T5DXNZPD5YAHHVIKVOE2PM@testcase.app.tld

       'P' for 'post' - locale - thread id - user id - signature digest
    """
    locale = get_locale()
    uid = "-".join(["P", str(locale), str(post.thread_id), str(member.id)])
    local_part = build_local_part(name, uid)
    return f"{local_part}@{domain}"


def extract_email_destination(address: str) -> tuple[str, ...]:
    """Return the values encoded in the email address.

    :param address: similar to test+IjEvMy8yLzQi.xjE04-4S0IzsdicTHKTAqcqa1fE@testcase.app.tld
    :return: List() of splitted values
    """
    m = re.search("<(.*)>", address)
    if m:
        address = m.group(1)
    local_part = address.rsplit("@", 1)[0]
    name, ident = local_part.rsplit("+", 1)
    uid, digest = ident.rsplit("-", 1)
    signed_local_part = build_local_part(name, uid)

    if local_part != signed_local_part:
        raise ValueError("Invalid signature in reply address")

    values = uid.split("-")
    header = values.pop(0)
    assert header == "P"
    return tuple(values)


def has_subtag(address: str) -> bool:
    """Return True if a subtag (delimited by '+') was found
    in the name part of the address.

    :param address: email adress
    """
    name = address.rsplit("@", 1)[0]
    return "+" in name


def send_post_to_user(community, post, member):
    """Send a post though SMTP.

    Configuring Flask-Mail
    Flask-Mail is configured through the standard Flask config API.
    These are the available options:

    MAIL_SERVER : default 'localhost'
    MAIL_PORT : default 25
    MAIL_USE_TLS : default False
    MAIL_USE_SSL : default False
    MAIL_DEBUG : default app.debug
    MAIL_USERNAME : default None
    MAIL_PASSWORD : default None
    MAIL_DEFAULT_SENDER : default None
    MAIL_MAX_EMAILS : default None
    MAIL_SUPPRESS_SEND : default app.testing
    MAIL_ASCII_ATTACHMENTS : default False
    """

    message = _mail_from_post(community, post, member)

    logger.info(f"Sending new post by email to {message.recipients}")
    try:
        # with mail.connect() as connection:
        with current_app.app_context():
            mail.send(message)
    except BaseException as e:
        # log to sentry if enabled
        logger.error(f"Send mail to user failed: {e}")


def _mail_from_post(community, post, member) -> Message:
    """Return a mail.Message build from a post in community forum"""
    recipient = member.email
    subject = f"[{community.name}] {post.title}"
    logger.info(f"{subject=} {recipient=}")

    config = current_app.config
    SENDER = config.get("BULK_MAIL_SENDER", config["MAIL_SENDER"])
    SBE_FORUM_REPLY_BY_MAIL = config.get("SBE_FORUM_REPLY_BY_MAIL", False)
    SBE_FORUM_REPLY_ADDRESS = config.get("SBE_FORUM_REPLY_ADDRESS", SENDER)
    SERVER_NAME = config.get("SERVER_NAME", "example.com")

    list_id = f'"{community.name} forum" <forum.{community.slug}.{SERVER_NAME}>'
    forum_url = url_for("forum.index", community_id=community.slug, _external=True)
    forum_archive_url = url_for(
        "forum.archives", community_id=community.slug, _external=True
    )
    extra_headers = {
        "List-Id": list_id,
        "List-Archive": f"<{forum_archive_url}>",
        "List-Post": f"<{forum_url}>",
        "X-Auto-Response-Suppress": "All",
        "Auto-Submitted": "auto-generated",
    }

    if SBE_FORUM_REPLY_BY_MAIL:
        name = SBE_FORUM_REPLY_ADDRESS.rsplit("@", 1)[0]
        domain = SBE_FORUM_REPLY_ADDRESS.rsplit("@", 1)[1]
        reply_to = build_reply_email_address(name, post, member, domain)
        sender = SBE_FORUM_REPLY_ADDRESS
    else:
        sender = SENDER
        reply_to = None

    msg = Message(
        subject,
        recipients=[recipient],
        sender=sender,
        reply_to=reply_to,
        extra_headers=extra_headers,
    )

    ctx = {
        "community": community,
        "post": post,
        "member": member,
        "MAIL_REPLY_MARKER": MAIL_REPLY_MARKER,
        "SBE_FORUM_REPLY_BY_MAIL": SBE_FORUM_REPLY_BY_MAIL,
    }
    msg.html = render_template_i18n("forum/mail/new_message.html", **ctx)
    msg.body = html2text.html2text(msg.html)
    return msg


def extract_content(payload, marker):
    """Search the payload for marker, return content up to marker."""
    index = payload.rfind(marker)
    content = payload[:index]
    return content


def validate_html(payload):
    css_sanitizer = CSSSanitizer(allowed_css_properties=ALLOWED_STYLES)
    args = {
        "tags": ALLOWED_TAGS,
        "attributes": ALLOWED_ATTRIBUTES,
        "css_sanitizer": css_sanitizer,
        "strip": True,
    }
    return bleach.clean(payload, **args).strip()


def add_paragraph(newpost):
    """Add surrounding <p>newpost</p> if necessary."""
    newpost = newpost.strip()
    if not newpost.startswith("<p>"):
        newpost = f"<p>{newpost}</p>"
    return newpost


def clean_html(newpost: str) -> str:
    """Clean leftover empty blockquotes."""

    clean = re.sub(
        r"(<blockquote.*?<p>.*?</p>.*?</blockquote>)",
        "",
        newpost,
        flags=re.MULTILINE | re.DOTALL,
    )

    # this cleans auto generated reponse text (<br>timedate <a email /a><br>)
    # we reverse the string because re.sub replaces
    #  LEFTMOST NON-OVERLAPPING OCCURRENCES, and we only want the last match
    # in the string
    clean = re.sub(
        r"(>rb<.*?>a.*?=ferh\sa<.*?>rb<)",
        "",
        clean[::-1],
        flags=re.MULTILINE | re.DOTALL,
    )
    clean = clean[::-1]

    return clean


def decode_payload(part: email.message.Message) -> str:
    """Get the payload and decode (base64 & quoted printable)."""

    payload = part.get_payload(decode=True)

    if isinstance(payload, str):
        return payload

    # Please the typechecker (and make things a bit clearer)
    assert isinstance(payload, bytes)
    payload_bytes: bytes = payload

    charset = part.get_content_charset()
    if charset is not None:
        try:
            payload_str = payload_bytes.decode(charset)
        except UnicodeDecodeError:
            payload_str = payload_bytes.decode("raw-unicode-escape")
    else:
        # What about other encodings? -> using chardet
        found = chardet.detect(payload_bytes)
        payload_str = payload_bytes.decode(found["encoding"])

    return payload_str


def process(message: email.message.Message, marker: str) -> tuple[str, list[dict]]:
    """Check the message for marker presence and return the text up to it if
    present.

    :raises LookupError otherwise.
    :return: sanitized html upto marker from message and attachements
    """
    assert isinstance(message, email.message.Message)
    content = {"plain": "", "html": ""}
    attachments: list[dict[str, Any]] = []

    # Iterate all message's parts for text/*
    for part in message.walk():
        content_type = part.get_content_type()
        content_disposition = part.get("Content-Disposition")

        if content_disposition not in [None, "inline"]:
            attachments.append(
                {
                    "filename": part.get_filename(),
                    "content_type": part.get_content_type(),
                    "data": part.get_payload(decode=True),
                }
            )

        elif content_type in ["text/plain", "text/html"]:
            subtype = part.get_content_subtype()
            payload = decode_payload(part)
            content[subtype] += payload

    if marker in content["html"]:
        newpost = extract_content(content["html"], marker[:9])
        newpost = add_paragraph(validate_html(newpost))
        newpost = clean_html(newpost)
    elif marker in content["plain"]:
        newpost = extract_content(content["plain"], marker[:9])
        newpost = add_paragraph(newpost)
    elif content["html"]:
        newpost = content["html"]
        newpost = add_paragraph(validate_html(newpost))
        newpost = clean_html(newpost)
    else:
        newpost = content["plain"]
        newpost = add_paragraph(newpost)
    # else:
    #     raise LookupError(f"No marker:{marker} in email")

    return newpost, attachments


@dramatiq.actor(max_retries=20, max_backoff=86400000)
def process_email(message: email.message.Message) -> bool:
    """Email.Message object from command line script Run message (parsed
    email).

    Processing chain extract community thread post member from reply_to
    persist post in db.
    """
    app = unwrap(current_app)
    # Extract post destination from To: field, (community/forum/thread/member)
    to_address = message["To"]

    assert isinstance(to_address, str)

    if not (has_subtag(to_address)):
        logger.info("Email %r has no subtag, skipping...", to_address)
        return False

    try:
        infos = extract_email_destination(to_address)
        locale = infos[0]
        thread_id = infos[1]
        user_id = infos[2]
    except BaseException:
        logger.error(
            "Recipient %r cannot be converted to locale/thread_id/user.id",
            to_address,
            exc_info=True,
        )
        return False

    # Translate marker with locale from email address
    rq_headers = [("Accept-Language", locale)]
    with app.test_request_context("/process_email", headers=rq_headers):
        marker = str(MAIL_REPLY_MARKER)

    # Extract text and attachments from message
    try:
        newpost, attachments = process(message, marker)
    except BaseException:
        logger.error("Could not Process message", exc_info=True)
        return False

    # Persist post
    with current_app.test_request_context("/process_email", headers=rq_headers):
        g.user = User.query.get(user_id)
        thread = Thread.query.get(thread_id)
        community = thread.community
        # FIXME: check membership, send back an informative email in case of an
        # error
        post = thread.create_post(body_html=newpost)
        obj_meta = post.meta.setdefault("abilian.sbe.forum", {})
        obj_meta["origin"] = "email"
        obj_meta["send_by_email"] = True
        activity.send(app, actor=g.user, verb="post", object=post, target=community)

        for desc in attachments:
            filename = desc["filename"]
            content_type = desc["content_type"]
            data = desc["data"]
            if isinstance(data, str):
                data = data.encode("utf8")
            attachment = PostAttachment(name=filename)
            attachment.post = post
            attachment.set_content(data, content_type=content_type)
            db.session.add(attachment)

        db.session.commit()

    # Notify all parties involved
    send_post_by_email.send(post.id)
    return True


@crontab("SCHEDULE_CHECK_MAILDIR")
@dramatiq.actor(max_retries=20, max_backoff=86400000)
def check_maildir():
    """Check the MailDir for emails to be injected in Threads.

    This task is registered only if `INCOMING_MAIL_USE_MAILDIR` is True.
    By default it is run every minute.
    """
    home = expanduser("~")
    maildirpath = Path(home) / "Maildir"
    if not maildirpath.is_dir():
        raise ValueError(f"{maildirpath} must be a directory")

    incoming_mailbox = mailbox.Maildir(str(maildirpath))
    incoming_mailbox.lock()  # Useless but recommended if old mbox is used by error

    try:
        for key, message in incoming_mailbox.items():
            processed = process_email(message)

            # delete the message if all went fine
            if processed:
                del incoming_mailbox[key]

    finally:
        incoming_mailbox.close()  # Flushes all changes to disk then unlocks



from __future__ import annotations

import sys
from pathlib import Path
from typing import IO

import pytest
from sqlalchemy.orm import Session

from abilian.core.models.subjects import User
from abilian.sbe.app import Application
from abilian.sbe.apps.communities.models import Community
from abilian.sbe.apps.documents.models import Document, Folder
from abilian.sbe.apps.documents.views.folders import explore_archive
from abilian.services import index_service, security_service
from abilian.testing.util import login, redis_available


def open_file(filename: str) -> IO[bytes]:
    path = Path(__file__).parent / "data" / "dummy_files" / filename
    return path.open("rb")


@pytest.mark.skipif(not redis_available(), reason="requires redis connection")
def test_document(app: Application, session: Session) -> None:
    root = Folder(title="root")
    doc = Document(parent=root, title="test")
    data = open_file("onepage.pdf").read()
    doc.set_content(data, "application/pdf")
    session.add(doc)
    session.commit()

    # coverage
    app.config["ANTIVIRUS_CHECK_REQUIRED"] = True
    doc.ensure_antivirus_scheduled()


def test_antivirus_properties(app: Application, session: Session) -> None:
    root = Folder(title="root")
    doc = Document(parent=root, title="test")
    doc.set_content(b"content", "text/plain")
    appcfg = app.config
    meta = doc.content_blob.meta

    # 1: not check required ################
    # no data in meta
    appcfg["ANTIVIRUS_CHECK_REQUIRED"] = False
    assert doc.antivirus_scanned is False
    assert doc.antivirus_status is None
    assert doc.antivirus_required is False
    assert doc.antivirus_ok is True

    # antivirus was run, but no result
    meta["antivirus"] = None
    assert doc.antivirus_scanned is True
    assert doc.antivirus_status is None
    assert doc.antivirus_required is False
    assert doc.antivirus_ok is True

    # virus detected
    meta["antivirus"] = False
    assert doc.antivirus_scanned is True
    assert doc.antivirus_status is False
    assert doc.antivirus_required is False
    assert doc.antivirus_ok is False

    # virus free
    meta["antivirus"] = True
    assert doc.antivirus_scanned is True
    assert doc.antivirus_status is True
    assert doc.antivirus_required is False
    assert doc.antivirus_ok is True

    # 2: check required ##################
    # no data in meta
    del meta["antivirus"]
    appcfg["ANTIVIRUS_CHECK_REQUIRED"] = True
    assert doc.antivirus_scanned is False
    assert doc.antivirus_status is None
    assert doc.antivirus_required is True
    assert doc.antivirus_ok is False

    # antivirus was run, but no result
    meta["antivirus"] = None
    assert doc.antivirus_scanned is True
    assert doc.antivirus_status is None
    assert doc.antivirus_required is True
    assert doc.antivirus_ok is False

    # virus detected
    meta["antivirus"] = False
    assert doc.antivirus_scanned is True
    assert doc.antivirus_status is False
    assert doc.antivirus_required is False
    assert doc.antivirus_ok is False

    # virus free
    meta["antivirus"] = True
    assert doc.antivirus_scanned is True
    assert doc.antivirus_status is True
    assert doc.antivirus_required is False
    assert doc.antivirus_ok is True


def test_folder_indexed(
    app: Application,
    session: Session,
    community1: Community,
    community2: Community,
    monkeypatch,
) -> None:
    monkeypatch.setenv("TESTING_DIRECT_FUNCTION_CALL", "testing")
    index_service.start()
    security_service.start()

    folder = Folder(title="Folder 1", parent=community1.folder)
    session.add(folder)

    folder_other = Folder(title="Folder 2: other", parent=community2.folder)
    session.add(folder_other)

    user_no_community = User(email="no_community@example.com", can_login=True)
    session.add(user_no_community)

    session.commit()

    svc = index_service
    obj_types = (Folder.entity_type,)

    with login(user_no_community):
        res = svc.search("folder", object_types=obj_types)
        assert len(res) == 0

    with login(community1.test_user):
        res = svc.search("folder", object_types=obj_types)
        assert len(res) == 1
        hit = res[0]
        assert hit["object_key"] == folder.object_key
        assert hit["community_slug"] == community1.slug

    with login(community2.test_user):
        res = svc.search("folder", object_types=obj_types)
        assert len(res) == 1
        hit = res[0]
        assert hit["object_key"] == folder_other.object_key
        assert hit["community_slug"] == community2.slug


@pytest.mark.skipif(sys.version_info >= (3, 0), reason="Doesn't work yet on Py3k")
def test_explore_archive():
    fd = open_file("content.zip")
    result = [("/".join(path), f) for path, f in explore_archive(fd)]
    assert result == [("", fd)]

    fd = open_file("content.zip")
    archive_content = explore_archive(fd, uncompress=True)
    result = {f"{'/'.join(path)}/{f.filename}" for path, f in archive_content}
    assert result == {
        "existing-doc/file.txt",
        "existing-doc/subfolder_in_renamed/doc.txt",
        "folder 1/doc.txt",
        "folder 1/dos cp437: é.txt",
        # FIXME: there is currently a regression with the way unicode filenames are decoded.
        # See: https://stackoverflow.com/questions/13261347/
        # (On MacOS, the utf-8 bit is not set by the Compress utility, and zipfile
        # incorreclty assumes CP437).
        # "folder 1/osx: utf-8: é.txt",
    }



from __future__ import annotations

from collections import namedtuple

from apscheduler.schedulers.blocking import BlockingScheduler
from apscheduler.triggers.cron import CronTrigger
from loguru import logger

Job = namedtuple("Job", ["actor", "key"])

DEFAULT_SCHEDULE = {
    "SCHEDULE_SEND_DAILY_SOCIAL_DIGEST": "0 10 * * *",
    "PERIODIC_CLEAN_UPLOAD_DIRECTORY": "0 * * * *",
    "SCHEDULE_CHECK_MAILDIR": "* * * * *",
}

_actor_registry = set()


def crontab(crontab: str):
    def decorator(func):
        logger.debug(f"Registering cron job: {func}")
        job = Job(actor=func, key=crontab)
        _actor_registry.add(job)
        return func

    return decorator


def crontab_from_config(config: dict, job: Job) -> str:
    key = job.key.upper()
    if key in config:
        # schedule value can be ""
        schedule = config[key].strip()
    else:
        schedule = DEFAULT_SCHEDULE.get(key, "")
    return schedule


def run_scheduler(config: dict):
    scheduler = BlockingScheduler()
    scheduler.remove_all_jobs()

    for job in _actor_registry:
        crontab_content = crontab_from_config(config, job)
        if not crontab_content:
            logger.info(f"NOT registering cron job: {job.actor}")
            continue
        logger.info(f"Registering cron job: {job.actor} {crontab_content}")

        scheduler.add_job(
            job.actor.send,
            CronTrigger.from_crontab(crontab_content),
        )

    try:
        scheduler.start()
    except KeyboardInterrupt:
        scheduler.shutdown()

    return 0



from __future__ import annotations

from datetime import datetime

from flask import url_for
from pytest import mark

from abilian.sbe.apps.calendar.models import Event


def test_create_event() -> None:
    start = datetime.now()
    event = Event(name="Test thread", start=start)
    assert event  # TODO


@mark.skip
def test(community1, client):
    response = client.get(url_for("calendar.index", community_id=community1.slug))
    assert response.status_code == 200


# @mark.skip
# def test_event_indexed(community1, community2, db, client, req_ctx):
#     start = datetime.now()
#     event1 = Event(name="Test event", community=community1, start=start)
#     event2 = Event(name="Test other event", community=community2, start=start)
#     db.session.add(event1)
#     db.session.add(event2)
#     db.session.commit()
#
#     svc = self.svc
#     obj_types = (Event.entity_type,)
#     with self.login(self.user_no_community):
#         res = svc.search("event", object_types=obj_types)
#         assert len(res) == 0
#
#     with self.login(self.user):
#         res = svc.search("event", object_types=obj_types)
#         assert len(res) == 1
#         hit = res[0]
#         assert hit["object_key"] == event1.object_key
#
#     with self.login(self.user_c2):
#         res = svc.search("event", object_types=obj_types)
#         assert len(res) == 1
#         hit = res[0]
#         assert hit["object_key"] == event2.object_key



""""""
from __future__ import annotations

from flask import Flask

from .criterion import TagCriterion
from .extension import TagsExtension

__all__ = ["TagCriterion", "TagsExtension"]


def register_plugin(app: Flask):
    TagsExtension(app)



""""""
from __future__ import annotations

from collections.abc import Callable

from flask import render_template

from abilian.i18n import _, _l
from abilian.web.admin.panel import AdminPanel
from abilian.web.util import url_for

from . import views


class GroupsPanel(AdminPanel):
    """Group administration panel."""

    id = "groups"
    label = _l("Groups")
    icon = "grain"

    def install_additional_rules(self, add_url_rule: Callable):
        add_url_rule("/groups", view_func=views.JsonGroupsList.as_view("json_list"))
        add_url_rule("/new", view_func=views.GroupCreate.as_view("new"))
        add_url_rule("/<int:group_id>/", view_func=views.GroupView.as_view("group"))
        add_url_rule(
            "/<int:group_id>/edit",
            view_func=views.GroupEdit.as_view(
                "group_edit", view_endpoint=".groups_group"
            ),
        )

    def get(self):
        # FIXME: use widgets.AjaxMainTableView instead
        datatable_options = {
            "sDom": "lfFrtip",
            "aaSorting": [[0, "asc"]],
            "aoColumns": [
                {"asSorting": ["asc", "desc"]},
                {"bSortable": False},
                {"bSortable": False},
            ],
            "bFilter": True,
            "oLanguage": {
                "sSearch": _("Filter records:"),
                "sPrevious": _("Previous"),
                "sNext": _("Next"),
                "sInfo": _("Showing _START_ to _END_ of _TOTAL_ entries"),
                "sInfoFiltered": _("(filtered from _MAX_ total entries)"),
                "sAddAdvancedFilter": _("Add a filter"),
            },
            "bStateSave": False,
            "bPaginate": True,
            "sPaginationType": "bootstrap",
            "bLengthChange": False,
            "iDisplayLength": 30,
            "bProcessing": True,
            "bServerSide": True,
            "sAjaxSource": url_for(".groups_json_list"),
        }

        return render_template(
            "admin/groups.html", next=next, datatable_options=datatable_options
        )



""""""
from __future__ import annotations

from collections.abc import Callable, Collection
from typing import Any, cast

from flask import Blueprint as BaseBlueprint

from abilian.core.models.subjects import User
from abilian.services.security import Anonymous, Role
from abilian.services.security.service import SecurityService


def allow_anonymous(user: User, roles: Collection[Role], **kwargs: Any) -> bool:
    return True


def allow_access_for_roles(roles: Collection[Role] | Role) -> Callable:
    """Access control helper to check user's roles against a list of valid
    roles."""
    if isinstance(roles, Role):
        roles = (roles,)
    valid_roles = frozenset(roles)

    if Anonymous in valid_roles:
        return allow_anonymous

    # FIXME: parameter not used. Why?
    def check_role(user: User, roles: Collection[Role], **kwargs):
        from abilian.services import get_service

        security: SecurityService = cast(SecurityService, get_service("security"))
        return security.has_role(user, valid_roles)

    return check_role


class Blueprint(BaseBlueprint):
    """An enhanced :class:`flask.blueprints.Blueprint` with access control
    helpers."""

    def __init__(
        self,
        name: str,
        import_name: str,
        allowed_roles: None | str | Role | Collection[Role] = None,
        **kwargs: Any,
    ):
        """
        :param allowed_roles: role or list of roles required to access any view in this
            blueprint.
        """
        super().__init__(name, import_name, **kwargs)

        if allowed_roles is not None:
            if isinstance(allowed_roles, str):
                allowed_roles = Role(allowed_roles)

            if isinstance(allowed_roles, Role):
                allowed_roles = (allowed_roles,)
        else:
            allowed_roles = ()

        if allowed_roles:
            self.record_once(
                lambda s: s.app.add_access_controller(
                    self.name, allow_access_for_roles(allowed_roles)
                )
            )

    def allow_any(self, func):
        pass



from __future__ import annotations

import logging
import sys

from flask import Flask
from loguru import logger


def init_logging(app: Flask) -> None:
    init_loguru(app)


class InterceptHandler(logging.Handler):
    def emit(self, record):
        logger_opt = logger.opt(depth=6, exception=record.exc_info)
        logger_opt.log(record.levelno, record.getMessage())


def init_loguru(app: Flask):
    # register loguru as (sole) handler
    # level = app.config.get('LOG_LEVEL', 'INFO')
    if app.debug:
        level = "DEBUG"
    else:
        level = "INFO"

    set_loguru_config(level)

    app.logger.handlers = []
    app.logger.addHandler(InterceptHandler())
    logger.info("Loguru initialized")


def set_loguru_config(level: str = "DEBUG"):
    logger.remove()
    logger.add(
        sys.stderr,
        level=level,
        backtrace=True,
        diagnose=True,
        enqueue=True,
        catch=True,
    )



from __future__ import annotations

from pathlib import Path

from pytest import fixture

from .. import CROP, SCALE, get_save_format, get_size, resize


@fixture()
def orig_image() -> bytes:
    # 725x518
    return (Path(__file__).parent / "cat.jpg").open("rb").read()


def test_get_save_format():
    assert get_save_format("JPG") == "JPEG"
    assert get_save_format("JPEG") == "JPEG"
    assert get_save_format("PNG") == "PNG"
    assert get_save_format("GIF") == "PNG"
    assert get_save_format("unknown") == "JPEG"


def test_fit(orig_image: bytes):
    image = resize(orig_image, 500, 500)
    x, y = get_size(image)
    assert (x, y) == (500, 357)

    # image already fits in desired dimension
    image = resize(orig_image, 1000, 1000)
    x, y = get_size(image)
    assert x == 725
    assert y == 518


def test_scale(orig_image: bytes):
    image = resize(orig_image, 500, 500, SCALE)
    assert get_size(image) == (500, 500)


def test_crop(orig_image: bytes):
    image = resize(orig_image, 500, 500, CROP)
    assert get_size(image) == (500, 500)



""""""
from __future__ import annotations

from flask import current_app
from flask.cli import AppGroup
from loguru import logger

config_commands = AppGroup("config")


@config_commands.command()
def show(only_path=False):
    """Show the current config."""
    infos = ["\n", f'Instance path: "{current_app.instance_path}"']

    logger.info("\n  ".join(infos))



"""User preference service.

Notes:

- Preferences are user-specific.
- For application settings use
  :class:`abilian.services.settings.SettingsService`.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any

from flask import Blueprint, Flask, g, redirect, request, url_for
from flask_login import current_user
from werkzeug.exceptions import InternalServerError

from abilian.core import signals
from abilian.core.extensions import db
from abilian.core.models.subjects import User
from abilian.i18n import _, _l
from abilian.services.auth.service import user_menu
from abilian.services.base import Service, ServiceState
from abilian.services.preferences.panel import PreferencePanel
from abilian.web.action import Endpoint
from abilian.web.nav import BreadcrumbItem, NavItem

from .models import UserPreference

if TYPE_CHECKING:
    from abilian.app import Application

_PREF_NAV_ITEM = NavItem(
    "user",
    "preferences",
    title=_l("Preferences"),
    icon="cog",
    url=lambda context: f"{request.url_root}preferences",
    condition=lambda context: not current_user.is_anonymous,
)

user_menu.insert(0, _PREF_NAV_ITEM)


class PreferenceState(ServiceState):
    panels: list[PreferencePanel]
    blueprint: Blueprint
    blueprint_registered: bool

    def __init__(self, service: PreferenceService, *args: Any, **kwargs: Any):
        super().__init__(service, *args, **kwargs)
        self.panels = []
        self.nav_paths = {}
        self.breadcrumb_items = {}
        self.blueprint_registered = False


class PreferenceService(Service):
    """Flask extension for a user-level preference service, with pluggable
    panels."""

    name = "preferences"
    AppStateClass = PreferenceState

    def init_app(self, app: Application, *panels: Any):
        super().init_app(app)

        with app.app_context():
            self.setup_blueprint(app)
            for panel in panels:
                self.register_panel(panel)

    def get_preferences(self, user: User | None = None) -> dict[str, Any]:
        """Return a string->value dictionnary representing the given user
        preferences.

        If no user is provided, the current user is used instead.
        """
        if user is None:
            user = current_user
        return {pref.key: pref.value for pref in user.preferences}

    def set_preferences(self, user: User | None = None, **kwargs: Any):
        """Set preferences from keyword arguments."""
        if user is None:
            user = current_user

        d = {pref.key: pref for pref in user.preferences}
        for k, v in kwargs.items():
            if k in d:
                d[k].value = v
            else:
                d[k] = UserPreference(user=user, key=k, value=v)
                db.session.add(d[k])

    def clear_preferences(self, user: User | None = None):
        """Clear the user preferences."""
        if user is None:
            user = current_user

        #  don't delete UserPreference 1 by 1 with session.delete, else
        #  user.preferences is not updated until commit() (not flush()). see
        #  http://docs.sqlalchemy.org/en/rel_0_7/orm/session.html#deleting-from-collections
        user.preferences = []

    def register_panel(self, panel: PreferencePanel, app: Flask | None = None):
        state = self.app_state if app is None else app.extensions[self.name]
        if state.blueprint_registered:
            raise ValueError(
                "Extension already initialized for app, cannot add more panel"
            )

        state.panels.append(panel)
        panel.preferences = self
        rule = f"/{getattr(panel, 'path', panel.id)}"
        endpoint = panel.id
        abs_endpoint = f"preferences.{endpoint}"

        if hasattr(panel, "get"):
            state.blueprint.add_url_rule(rule, endpoint, panel.get)
        if hasattr(panel, "post"):
            endpoint += "_post"
            state.blueprint.add_url_rule(rule, endpoint, panel.post, methods=["POST"])

        state.breadcrumb_items[abs_endpoint] = BreadcrumbItem(
            label=panel.label, icon=None, url=Endpoint(abs_endpoint)
        )

    def setup_blueprint(self, app: Flask):
        bp = self.app_state.blueprint = Blueprint(
            "preferences",
            __name__,
            template_folder="templates",
            url_prefix="/preferences",
        )

        # we need to delay blueprint registration to allow adding more panels during
        # initialization
        @signals.components_registered.connect_via(app)
        def register_bp(app: Flask):
            app.register_blueprint(bp)
            app.extensions[self.name].blueprint_registered = True

        self.app_state.root_breadcrumb_item = BreadcrumbItem(
            label=_("Preferences"), url=Endpoint("preferences.index")
        )

        bp.url_value_preprocessor(self.build_breadcrumbs)

        @bp.context_processor
        def inject_menu() -> dict[str, list[dict[str, Any]]]:
            menu = []
            for panel in self.app_state.panels:
                if not panel.is_accessible():
                    continue
                endpoint = f"preferences.{panel.id}"
                active = endpoint == request.endpoint
                entry = {
                    "endpoint": endpoint,
                    "label": panel.label,
                    "url": url_for(endpoint),
                    "active": active,
                }
                menu.append(entry)
            return {"menu": menu}

        @bp.route("/")
        def index():
            """Index redirects to the first accessible panel."""

            # Work around unit test failure. FIXME.
            if current_user.is_anonymous:
                return "OK"

            for panel in self.app_state.panels:
                if panel.is_accessible():
                    return redirect(url_for(f"preferences.{panel.id}"))

            # Should not happen.
            raise InternalServerError

    def build_breadcrumbs(self, endpoint, view_args):
        state = self.app_state
        g.nav["active"] = _PREF_NAV_ITEM.path
        g.breadcrumb.append(state.root_breadcrumb_item)
        if endpoint in state.breadcrumb_items:
            g.breadcrumb.append(state.breadcrumb_items[endpoint])


preferences = PreferenceService()



"""Subject classes (i.e. people, groups, etc.).

See ICOM-ics-v1.0 "Subject Branch".

TODO: I'm not a big fan of the "subject" name. Could be replaced by something
else, like "people" or "principal" ?
"""
from __future__ import annotations

import random
import string
from abc import ABCMeta, abstractmethod
from datetime import datetime, timedelta

import bcrypt
import sqlalchemy as sa
from flask_login import UserMixin
from flask_sqlalchemy import BaseQuery
from sqlalchemy.event import listens_for
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import backref, deferred, relationship
from sqlalchemy.orm.mapper import Mapper
from sqlalchemy.schema import Column, ForeignKey, UniqueConstraint
from sqlalchemy.types import Boolean, DateTime, Integer, LargeBinary, UnicodeText

from abilian.core import sqlalchemy as sa_types
from abilian.core.util import fqcn

from .base import SEARCHABLE, SYSTEM, IdMixin, Indexable, TimestampedMixin, db

__all__ = (
    "ClearPasswordStrategy",
    "Group",
    "Principal",
    "User",
    "create_root_user",
    "gen_random_password",
)

# Tables for many-to-many relationships
following = db.Table(
    "following",
    db.Model.metadata,
    Column("follower_id", Integer, ForeignKey("user.id")),
    Column("followee_id", Integer, ForeignKey("user.id")),
    UniqueConstraint("follower_id", "followee_id"),
)

membership = db.Table(
    "membership",
    db.Model.metadata,
    Column(
        "user_id",
        Integer,
        ForeignKey("user.id", onupdate="CASCADE", ondelete="CASCADE"),
    ),
    Column(
        "group_id",
        Integer,
        ForeignKey("group.id", onupdate="CASCADE", ondelete="CASCADE"),
    ),
    UniqueConstraint("user_id", "group_id"),
)

# Should not be needed (?)
administratorship = db.Table(
    "administratorship",
    db.Model.metadata,
    Column(
        "user_id",
        Integer,
        ForeignKey("user.id", onupdate="CASCADE", ondelete="CASCADE"),
    ),
    Column(
        "group_id",
        Integer,
        ForeignKey("group.id", onupdate="CASCADE", ondelete="CASCADE"),
    ),
    UniqueConstraint("user_id", "group_id"),
)

_RANDOM_PASSWORD_CHARS = string.ascii_letters + string.digits + string.punctuation


def gen_random_password(length=15):
    rg = random.SystemRandom()
    return "".join(rg.choice(_RANDOM_PASSWORD_CHARS) for i in range(length))


class PasswordStrategy(metaclass=ABCMeta):
    @property
    @abstractmethod
    def name(self):
        """Strategy name."""

    @abstractmethod
    def authenticate(self, user, password):
        """Predicate to tell wether password match user's or not."""

    @abstractmethod
    def process(self, user, password):
        """Return a string to be stored as user password."""


class ClearPasswordStrategy(PasswordStrategy):
    """Don't encrypt at all.

    This strategy should not ever be used elsewhere than in tests. It's
    useful in tests since a hash like bcrypt is designed to be slow.
    """

    @property
    def name(self):
        return "clear"

    def authenticate(self, user, password):
        return user.password == password

    def process(self, user, password):
        if not isinstance(password, str):
            password = password.decode("utf-8")
        return password


class BcryptPasswordStrategy(PasswordStrategy):
    """Hash passwords using bcrypt."""

    @property
    def name(self):
        return "bcrypt"

    def authenticate(self, user: User, password: str) -> bool:
        assert isinstance(password, str)

        current_passwd = user.password
        # crypt work only on bytes, not str (Unicode)
        if isinstance(current_passwd, str):
            current_passwd = current_passwd.encode("utf-8")
        if isinstance(password, str):
            password = password.encode("utf-8")

        return bcrypt.hashpw(password, current_passwd) == current_passwd

    def process(self, user: User, password: str) -> str:
        if isinstance(password, str):
            password = password.encode("utf-8")
        return bcrypt.hashpw(password, bcrypt.gensalt()).decode("utf-8")


class UserQuery(BaseQuery):
    def get_by_email(self, email):
        return self.filter_by(email=email).one()


class Principal(IdMixin, TimestampedMixin, Indexable):
    """A principal is either a User or a Group."""

    __index_to__ = (("name", ("name", "name_prefix", "text")),)

    query_class = UserQuery
    query: UserQuery

    def has_role(self, role, context=None):
        from abilian.services import get_security_service

        security_service = get_security_service()
        return security_service.has_role(self, role, context)


def set_entity_type(cls: type[User | Group]) -> type[User | Group]:
    """Decorator used to set the class' entity_type after the class has been
    declared.

    Actually, it works using __module__ during class declaration, but
    linters (Flake8, PyCharm) complain.
    """
    cls.entity_type = fqcn(cls)
    return cls


@set_entity_type
class User(Principal, UserMixin, db.Model):
    __tablename__ = "user"
    __editable__ = ["first_name", "last_name", "email", "password"]
    __exportable__ = __editable__ + ["created_at", "updated_at", "id"]

    __password_strategy__ = BcryptPasswordStrategy()

    # Basic information
    first_name = Column(UnicodeText, info=SEARCHABLE)
    last_name = Column(UnicodeText, info=SEARCHABLE)
    # Should we add gender, salutation ?

    # System information
    email = Column(UnicodeText, nullable=False)
    can_login = Column(Boolean, nullable=False, default=True)
    password = Column(UnicodeText, default="*", info={"audit_hide_content": True})

    photo = deferred(Column(LargeBinary))

    last_active = Column(DateTime, info=SYSTEM)
    locale = Column(sa_types.Locale, nullable=True, default=None)
    timezone = Column(sa_types.Timezone, nullable=True, default=None)

    __table_args__ = (UniqueConstraint("email"),)

    followers = relationship(
        "User",
        secondary=following,
        primaryjoin="User.id == following.c.follower_id",
        secondaryjoin="User.id == following.c.followee_id",
        backref="followees",
    )

    def __init__(self, password=None, **kwargs):
        Principal.__init__(self)
        UserMixin.__init__(self)
        db.Model.__init__(self, **kwargs)

        if self.can_login and password is not None:
            self.set_password(password)
            self._password = password

    def authenticate(self, password: str) -> bool:
        if self.password and self.password != "*":  # noqa: S105
            return self.__password_strategy__.authenticate(self, password)
        else:
            return False

    def set_password(self, password: str):
        """Encrypts and sets password."""
        self.password = self.__password_strategy__.process(self, password)

    def follow(self, followee):
        if followee == self:
            raise Exception("User can't follow self")
        self.followees.append(followee)

    def unfollow(self, followee):
        if followee == self:
            raise Exception("User can't follow self")
        i = self.followees.index(followee)
        del self.followees[i]

    def join(self, group):
        self.groups.add(group)

    def leave(self, group):
        if group in self.groups:
            self.groups.remove(group)

    #
    # Boolean properties
    #
    def is_following(self, other):
        return other in self.followees

    def is_member_of(self, group):
        return self in group.members

    def is_admin_of(self, group):
        return self in group.admins

    @property
    def is_online(self):
        if self.last_active is None:
            return False
        return datetime.utcnow() - self.last_active <= timedelta(minutes=1)

    #
    # Other properties
    #
    @property
    def name(self) -> str:
        name = f"{(self.first_name or '')} {(self.last_name or '')}"
        return name.strip() or "Unknown"

    @property
    def short_name(self):
        first_name = self.first_name or ""
        last_name = f"{self.last_name[0:1]}." if self.last_name else ""
        name = f"{first_name} {last_name}"
        return name.strip() or "Unknown"

    def __str__(self) -> str:
        return self.name

    def __repr__(self):
        cls = self.__class__
        return "<{mod}.{cls} id={id!r} email={email!r} at 0x{addr:x}>".format(
            mod=cls.__module__,
            cls=cls.__name__,
            id=self.id,
            email=self.email,
            addr=id(self),
        )


@listens_for(User, "mapper_configured", propagate=True)
def _add_user_indexes(mapper: Mapper, cls: type[User]):
    # this is a functional index (indexes on a function result), we cannot define
    # it in __table_args__.
    #
    # see: https://groups.google.com/d/msg/sqlalchemy/CgSJUlelhGs/_Nj3f201hs4J
    idx = sa.schema.Index(
        "user_unique_lowercase_email", sa.sql.func.lower(cls.email), unique=True
    )
    idx.info["engines"] = ("postgresql",)


@set_entity_type
class Group(Principal, db.Model):
    __indexable__ = False
    __tablename__ = "group"
    __editable__ = ["name", "description"]
    __exportable__ = __editable__ + ["created_at", "updated_at", "id"]

    name = Column(UnicodeText, nullable=False, info=SEARCHABLE)
    description = Column(UnicodeText, info=SEARCHABLE)
    photo = deferred(Column(LargeBinary))
    public = Column(Boolean, default=False, nullable=False)

    members = relationship(
        "User",
        collection_class=set,
        secondary=membership,
        backref=backref("groups", lazy="select", collection_class=set),
    )
    admins = relationship("User", collection_class=set, secondary=administratorship)

    @hybrid_property
    def members_count(self):
        return len(self.members)

    @members_count.expression
    def members_count(cls):
        return (
            sa.sql.select([sa.sql.func.count(membership.c.user_id)])
            .where(membership.c.group_id == cls.id)
            .group_by(membership.c.group_id)
            .label("members_count")
        )


def create_root_user() -> User:
    user = User.query.get(0)
    if user is None:
        user = User(
            id=0, last_name="SYSTEM", email="system@example.com", can_login=False
        )
        db.session.add(user)
        db.session.commit()
    return user



from __future__ import annotations

from tempfile import NamedTemporaryFile
from typing import IO

import pytest
from flask import g

from abilian.core.models.subjects import User
from abilian.core.sqlalchemy import SQLAlchemy
from abilian.sbe.apps.communities.models import READER, Community
from abilian.sbe.apps.communities.views.wizard import (
    wizard_extract_data,
    wizard_read_csv,
)


@pytest.fixture()
def csv_file() -> IO[str]:
    # create a tmp csv file
    csv = NamedTemporaryFile("w+", suffix=".csv", prefix="tmp_", delete=False)
    csv.write("user_1@example.com;userone;userone;manager\n")
    csv.write("user_2@example.com;usertwo;usertwo;member\n")
    csv.write("user_7@example.com;userseven;userseven;member\n")

    # writing a wrong line
    csv.write("user1@example.com;userthree;userthree\n")
    csv.write("example.com;example;userfour;member\n")

    csv.seek(0)

    return csv


def test_wizard_read_csv(csv_file: IO[str]) -> None:
    wizard_read = wizard_read_csv(csv_file)

    assert wizard_read == [
        {
            "first_name": "userone",
            "last_name": "userone",
            "role": "manager",
            "email": "user_1@example.com",
        },
        {
            "first_name": "usertwo",
            "last_name": "usertwo",
            "role": "member",
            "email": "user_2@example.com",
        },
        {
            "first_name": "userseven",
            "last_name": "userseven",
            "role": "member",
            "email": "user_7@example.com",
        },
    ]


def test_wizard_extract_data(db: SQLAlchemy, csv_file: IO[str]) -> None:
    session = db.session
    community = Community(name="Hp")
    g.community = community

    user1 = User(email="user_1@example.com")
    user2 = User(email="user_2@example.com")
    user3 = User(email="user_3@example.com")

    new_emails = [
        "user_1@example.com",
        "user_2@example.com",
        "user_3@example.com",
        "user_4@example.com",
        "user_5@example.com",
    ]

    # creating community
    session.add(community)

    # creating users
    session.add(user1)
    session.add(user2)
    session.add(user3)
    session.flush()

    # add user1 to the community
    community.set_membership(user1, READER)
    session.flush()

    # check wizard function in case of email list
    (
        existing_accounts_objects,
        existing_members_objects,
        accounts_list,
    ) = wizard_extract_data(new_emails)
    assert set(existing_accounts_objects) == {user2, user3}
    assert existing_members_objects == [user1]

    def sorter(x):
        return x["email"]

    assert sorted(accounts_list, key=sorter) == sorted(
        [
            {
                "status": "existing",
                "first_name": None,
                "last_name": None,
                "role": "member",
                "email": "user_2@example.com",
            },
            {
                "status": "existing",
                "first_name": None,
                "last_name": None,
                "role": "member",
                "email": "user_3@example.com",
            },
            {
                "status": "new",
                "first_name": "",
                "last_name": "",
                "role": "member",
                "email": "user_5@example.com",
            },
            {
                "status": "new",
                "first_name": "",
                "last_name": "",
                "role": "member",
                "email": "user_4@example.com",
            },
        ],
        key=sorter,
    )

    # check wizard function in case of csv file
    (
        existing_accounts_objects,
        existing_members_objects,
        accounts_list,
    ) = wizard_extract_data(csv_data=wizard_read_csv(csv_file))

    assert existing_accounts_objects == {
        "csv_roles": {
            "user_1@example.com": "manager",
            "user_2@example.com": "member",
            "user_7@example.com": "member",
        },
        "account_objects": [user2],
    }
    assert existing_members_objects == [user1]
    assert sorted(accounts_list, key=sorter) == sorted(
        [
            {
                "status": "existing",
                "first_name": None,
                "last_name": None,
                "role": "member",
                "email": "user_2@example.com",
            },
            {
                "status": "new",
                "first_name": "userseven",
                "last_name": "userseven",
                "role": "member",
                "email": "user_7@example.com",
            },
        ],
        key=sorter,
    )



""""""
from __future__ import annotations

from datetime import datetime
from functools import total_ordering

from flask_babel import LazyString
from flask_sqlalchemy import BaseQuery
from sqlalchemy import sql
from sqlalchemy.orm import backref, relationship
from sqlalchemy.schema import (
    CheckConstraint,
    Column,
    ForeignKey,
    Index,
    UniqueConstraint,
)
from sqlalchemy.types import Boolean, DateTime, Enum, Integer, String, UnicodeText

from abilian.core.entities import Entity
from abilian.core.extensions import db
from abilian.core.models import Model
from abilian.core.models.subjects import Group, User
from abilian.core.singleton import UniqueName, UniqueNameType
from abilian.i18n import _l

__all__ = [
    "CREATE",
    "DELETE",
    "MANAGE",
    "PERMISSIONS_ATTR",
    "READ",
    "WRITE",
    "Admin",
    "Anonymous",
    "Authenticated",
    "Creator",
    "FolderishModel",
    "InheritSecurity",
    "Manager",
    "Owner",
    "Permission",
    "PermissionAssignment",
    "Reader",
    "Role",
    "RoleAssignment",
    "RoleType",
    "SecurityAudit",
    "Writer",
]


@total_ordering
class Permission(UniqueName):
    """Defines permission by name.

    Permission instances are unique by name.
    """

    __slots__ = ("label",)

    def __init__(
        self, name: str, label: None | str | LazyString = None, assignable: bool = True
    ):
        super().__init__(name)
        if label is None:
            label = f"permission_{name!s}"
        if isinstance(label, str):
            label = _l(label)
        self.label = label

    def __str__(self) -> str:
        return str(self.name)

    def __lt__(self, other):
        return str(self.label).__lt__(str(other.label))


class PermissionType(UniqueNameType):
    """Store :class:`Permission`

    Usage:: RoleType()
    """

    Type = Permission


@total_ordering
class Role(UniqueName):
    """Defines role by name. Roles instances are unique by name.

    :param assignable: true if this role is can be assigned through security service.
        Non-assignable roles are roles automatically given depending on context (ex:
        Anonymous/Authenticated).
    """

    __slots__ = ("label", "assignable")

    def __init__(
        self, name: str, label: None | str | LazyString = None, assignable: bool = True
    ):
        super().__init__(name)
        if label is None:
            label = f"role_{name!s}"
        if isinstance(label, str):
            label = _l(label)
        self.label = label
        self.assignable = assignable

    def __str__(self) -> str:
        return str(self.name)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name}, {self.label})"

    def __lt__(self, other: Role) -> bool:
        return str(self.label).__lt__(str(other.label))

    @classmethod
    def assignable_roles(cls):
        return sorted(r for r in cls.__instances__.values() if r.assignable)


class RoleType(UniqueNameType):
    """Store :class:`Role`

    Usage:: RoleType()
    """

    Type = Role


#: marker for role assigned to 'Anonymous'
Anonymous = Role("anonymous", _l("role_anonymous"), assignable=False)

#: marker for role assigned to 'Authenticated'
Authenticated = Role("authenticated", _l("role_authenticated"), assignable=False)

#: marker for `admin` role
Admin = Role("admin", _l("role_administrator"))

#: marker for `manager` role
Manager = Role("manager", _l("role_manager"), assignable=False)

Creator = Role("creator", assignable=False)
Owner = Role("owner", assignable=False)
Reader = Role("reader", assignable=False)
Writer = Role("writer", assignable=False)

# Permissions
READ = Permission("read")
WRITE = Permission("write")
MANAGE = Permission("manage")
CREATE = Permission("create")
DELETE = Permission("delete")


class RoleAssignment(Model):
    __tablename__ = "roleassignment"
    __table_args__ = (
        #
        CheckConstraint(
            "(CAST(anonymous AS INTEGER) = 1)"
            " OR "
            "((CAST(anonymous AS INTEGER) = 0)"
            " AND "
            " ((user_id IS NOT NULL AND group_id IS NULL)"
            "  OR "
            "  (user_id IS NULL AND group_id IS NOT NULL)))",
            name="roleassignment_ck_user_xor_group",
        ),
        #
        UniqueConstraint(
            "anonymous",
            "user_id",
            "group_id",
            "role",
            "object_id",
            name="assignment_mapped_role_unique",
        ),
    )

    id = Column(Integer, primary_key=True, autoincrement=True, nullable=False)
    role = Column(RoleType, index=True, nullable=False)
    anonymous = Column(
        "anonymous",
        Boolean,
        index=True,
        nullable=True,
        default=False,
        server_default=sql.false(),
    )
    user_id = Column(Integer, ForeignKey("user.id", ondelete="CASCADE"), index=True)
    user = relationship(User, lazy="joined")
    group_id = Column(Integer, ForeignKey("group.id", ondelete="CASCADE"), index=True)
    group = relationship(Group, lazy="joined")

    object_id = Column(Integer, ForeignKey(Entity.id, ondelete="CASCADE"), index=True)
    object = relationship(Entity, lazy="select")


# On postgres the UniqueConstraint will not be effective because at least 1
# columns will have NULL value:
#
# From http://www.postgresql.org/docs/9.1/static/ddl-constraints.html
#
# "That means even in the presence of a unique constraint it is possible to
# store duplicate rows that contain a null value in at least one of the
# constrained columns."
#
# The solution is to build specific UNIQUE indexes, only for postgres
#
# noinspection PyComparisonWithNone
def _postgres_indexes1() -> list[Index]:
    role = RoleAssignment.role
    user_id = RoleAssignment.user_id
    group_id = RoleAssignment.group_id
    obj = RoleAssignment.object_id
    anonymous = RoleAssignment.anonymous
    name = "roleassignment_idx_{}_unique".format
    engines = ("postgresql",)
    indexes = [
        Index(
            name("user_role"),
            user_id,
            role,
            unique=True,
            postgresql_where=(
                (anonymous == False) & (group_id == None) & (obj == None)
            ),
        ),
        Index(
            name("group_role"),
            group_id,
            role,
            unique=True,
            postgresql_where=((anonymous == False) & (user_id == None) & (obj == None)),
        ),
        Index(
            name("anonymous_role"),
            role,
            unique=True,
            postgresql_where=(
                (anonymous == True)
                & (user_id == None)
                & (group_id == None)
                & (obj == None)
            ),
        ),
        Index(
            name("user_role_object"),
            user_id,
            role,
            obj,
            unique=True,
            postgresql_where=(
                (anonymous == False) & (group_id == None) & (obj != None)
            ),
        ),
        Index(
            name("group_role_object"),
            group_id,
            role,
            obj,
            unique=True,
            postgresql_where=((anonymous == False) & (user_id == None) & (obj != None)),
        ),
        Index(
            name("anonymous_role_object"),
            role,
            obj,
            unique=True,
            postgresql_where=(
                (anonymous == True)
                & (user_id == None)
                & (group_id == None)
                & (obj != None)
            ),
        ),
    ]

    for idx in indexes:
        idx.info["engines"] = engines

    return indexes


_postgres_indexes1()

PERMISSIONS_ATTR = "__permissions__"


class PermissionAssignment(db.Model):
    __tablename__ = "permission_assignment"
    __table_args__ = (
        UniqueConstraint("permission", "role", "object_id", name="assignments_unique"),
    )

    id = Column(Integer, primary_key=True, autoincrement=True, nullable=False)
    permission = Column(PermissionType, index=True, nullable=False)
    role = Column(RoleType, index=True, nullable=False)
    object_id = Column(
        Integer, ForeignKey(Entity.id, ondelete="CASCADE"), index=True, nullable=True
    )
    object = relationship(
        Entity,
        lazy="select",
        backref=backref(
            PERMISSIONS_ATTR,
            lazy="select",
            collection_class=set,
            cascade="all, delete-orphan",
            passive_deletes=True,
        ),
    )

    def __hash__(self) -> int:
        return hash((self.permission, self.role, self.object))

    def __eq__(self, other):
        if not isinstance(other, PermissionAssignment):
            return False

        return (
            self.permission == other.permission
            and self.role == other.role
            and self.object == other.object
        )

    def __neq__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        class_ = self.__class__
        classname = class_.__name__
        return (
            f"<{classname} instance at 0x{id(self):x} "
            f"permission={self.permission.name!r} "
            f"role={self.role.name!r} object={self.object!r}>"
            ""
        )


def _postgres_indexes2() -> list[Index]:
    # we need a unique index for when object_id is NULL; when it's not the
    # uniqueconstraint will just work.
    PA = PermissionAssignment
    name = "ix_permission_assignment_{}_unique".format
    engines = ("postgresql",)
    indexes = [
        Index(
            name("permission_role_global"),
            PA.permission,
            PA.role,
            unique=True,
            postgresql_where=(PA.object_id == None),
        )
    ]

    for idx in indexes:
        idx.info["engines"] = engines

    return indexes


_postgres_indexes2()


class SecurityAudit(db.Model):
    """Logs changes on security."""

    GRANT = "GRANT"
    REVOKE = "REVOKE"
    SET_INHERIT = "SET_INHERIT"
    UNSET_INHERIT = "UNSET_INHERIT"

    __tablename__ = "securityaudit"
    __table_args__ = (
        # constraint: either a inherit/no_inherit op on an object AND no user no group
        #             either a grant/revoke on a user XOR a group.
        CheckConstraint(
            f"(op IN ('{SET_INHERIT}', '{UNSET_INHERIT}') "
            " AND object_id IS NOT NULL"
            " AND user_id IS NULL "
            " AND group_id IS NULL "
            " AND (CAST(anonymous AS INTEGER) = 0)"
            ")"
            " OR "
            f"(op NOT IN ('{SET_INHERIT}', '{UNSET_INHERIT}')"
            " AND "
            " (((CAST(anonymous AS INTEGER) = 1) "
            "   AND user_id IS NULL AND group_id IS NULL)"
            "  OR "
            "  ((CAST(anonymous AS INTEGER) = 0) "
            "   AND ((user_id IS NOT NULL AND group_id IS NULL)"
            "  OR "
            "  (user_id IS NULL AND group_id IS NOT NULL)))"
            "))",
            name="securityaudit_ck_user_xor_group",
        ),
    )

    id = Column(Integer, primary_key=True)
    happened_at = Column(DateTime, default=datetime.utcnow, index=True)
    op = Column(
        Enum(GRANT, REVOKE, SET_INHERIT, UNSET_INHERIT, name="securityaudit_enum_op")
    )
    role = Column(RoleType)

    manager_id = Column(Integer, ForeignKey(User.id))
    manager = relationship(User, foreign_keys=manager_id)
    anonymous = Column("anonymous", Boolean, nullable=True, default=False)
    user_id = Column(Integer, ForeignKey("user.id"))
    user = relationship("User", lazy="joined", foreign_keys=user_id)
    group_id = Column(Integer, ForeignKey("group.id"))
    group = relationship("Group", lazy="joined")

    _fk_object_id = Column(Integer, ForeignKey(Entity.id, ondelete="SET NULL"))
    object = relationship(Entity, lazy="select")

    object_id = Column(Integer)
    object_type = Column(String(1000))
    object_name = Column(UnicodeText)

    query: BaseQuery


class InheritSecurity:
    """Mixin for objects with a parent relation and security inheritance."""

    inherit_security = Column(
        Boolean, default=True, nullable=False, info={"auditable": False}
    )


class FolderishModel(Entity, InheritSecurity):
    pass



from __future__ import annotations

import contextlib
import shutil
import typing
import weakref
from pathlib import Path
from typing import IO, Any
from uuid import UUID, uuid1

import sqlalchemy as sa
import sqlalchemy.event
from flask import _app_ctx_stack
from flask.globals import _lookup_app_object
from sqlalchemy.engine.base import Connection
from sqlalchemy.orm import scoped_session
from sqlalchemy.orm.session import Session, SessionTransaction
from sqlalchemy.orm.unitofwork import UOWTransaction

from abilian.core.extensions import db
from abilian.core.models import Model
from abilian.core.models.blob import Blob
from abilian.services import Service, ServiceState

if typing.TYPE_CHECKING:
    from abilian.app import Application

_NULL_MARK = object()


def _assert_uuid(uuid: Any):
    if not isinstance(uuid, UUID):
        raise TypeError("Not an uuid.UUID instance", uuid)


class BlobStoreServiceState(ServiceState):
    #: :class:`Path` path to application repository
    path: Path | None = None


class BlobStoreService(Service):
    """Service for storage of binary objects referenced in database."""

    name = "blob_store"
    AppStateClass = BlobStoreServiceState

    def init_app(self, app: Application):
        super().init_app(app)

        path = app.data_dir / "files"
        if not path.exists():
            path.mkdir(mode=0o775, parents=True)

        with app.app_context():
            self.app_state.path = path.resolve()

    # data management: paths and accessors
    def rel_path(self, uuid: UUID) -> Path:
        """Contruct relative path from blob store top directory to the file
        named after this uuid.

        :param:uuid: :class:`UUID` instance
        """
        _assert_uuid(uuid)

        filename = str(uuid)
        return Path(filename[0:2], filename[2:4], filename)

    def abs_path(self, uuid: UUID) -> Path:
        """Return absolute :class:`Path` object for given uuid.

        :param:uuid: :class:`UUID` instance
        """
        _assert_uuid(uuid)

        top = self.app_state.path
        rel_path = self.rel_path(uuid)
        dest = top / rel_path
        assert top in dest.parents
        return dest

    def get(self, uuid: UUID, default: Path | None = None) -> Path | None:
        """Return absolute :class:`Path` object for given uuid, if this uuid
        exists in blob store, or `default` if it doesn't.

        :param:uuid: :class:`UUID` instance
        """
        _assert_uuid(uuid)

        path = self.abs_path(uuid)
        if not path.exists():
            return default
        return path

    def set(self, uuid: UUID, content: Any, encoding: str | None = "utf-8"):
        """Store binary content with uuid as key.

        :param:uuid: :class:`UUID` instance
        :param:content: string, bytes, or any object with a `read()` method
        :param:encoding: encoding to use when content is Unicode
        """
        _assert_uuid(uuid)

        dest = self.abs_path(uuid)
        if not dest.parent.exists():
            dest.parent.mkdir(0o775, parents=True)

        if hasattr(content, "read"):
            content = content.read()

        mode = "tw"
        if not isinstance(content, str):
            mode = "bw"
            encoding = None

        with dest.open(mode, encoding=encoding) as f:
            f.write(content)

    def delete(self, uuid: UUID):
        """Delete file with given uuid.

        :param:uuid: :class:`UUID` instance
        :raises:KeyError if file does not exists
        """
        _assert_uuid(uuid)

        dest = self.abs_path(uuid)
        if not dest.exists():
            raise KeyError("No file can be found for this uuid", uuid)

        dest.unlink()

    def __getitem__(self, uuid: UUID) -> Path:
        _assert_uuid(uuid)

        value = self.get(uuid)
        if value is None:
            raise KeyError("No file can be found for this uuid", uuid)
        return value

    def __setitem__(self, uuid: UUID, content: Any):
        _assert_uuid(uuid)

        self.set(uuid, content)

    def __delitem__(self, uuid: UUID):
        _assert_uuid(uuid)

        self.delete(uuid)


blob_store = BlobStoreService()

_BLOB_STORE_TRANSACTION = "abilian_blob_store_transactions"


class SessionBlobStoreState(ServiceState):
    path: Path

    @property
    def transactions(self) -> dict[int, Any]:
        try:
            return _lookup_app_object(_BLOB_STORE_TRANSACTION)
        except AttributeError:
            reg: dict[int, Any] = {}
            setattr(_app_ctx_stack.top, _BLOB_STORE_TRANSACTION, reg)
            return reg

    @transactions.setter
    def transactions(self, value):
        top = _app_ctx_stack.top
        if top is None:
            raise RuntimeError("working outside of application context")

        setattr(top, _BLOB_STORE_TRANSACTION, value)

    # transaction <-> db session accessors
    def get_transaction(
        self, session: Session | scoped_session
    ) -> BlobStoreTransaction | None:
        if isinstance(session, scoped_session):
            session = session()

        s_id = id(session)
        default = (weakref.ref(session), None)
        s_ref, transaction = self.transactions.get(s_id, default)
        s = s_ref()

        if s is None or s is not session:
            # old session with same id, maybe not yet garbage collected
            transaction = None
        return transaction

    def set_transaction(
        self,
        session: Session | scoped_session,
        transaction: BlobStoreTransaction | None,
    ):
        if isinstance(session, scoped_session):
            session = session()

        session_id = id(session)
        self.transactions[session_id] = (weakref.ref(session), transaction)

    def create_transaction(self, session: Session, transaction: BlobStoreTransaction):
        if not self.running:
            return

        parent = self.get_transaction(session)
        root_path = self.path
        transaction = BlobStoreTransaction(root_path, parent)
        self.set_transaction(session, transaction)

    def end_transaction(self, session: Session, transaction: BlobStoreTransaction):
        if not self.running:
            return

        tr = self.get_transaction(session)
        if tr is not None:
            if not tr.cleared:
                # root and nested transactions emit "commit", but
                # subtransactions don't
                tr.commit(session)
            self.set_transaction(session, tr._parent)

    def begin(self, session: Session):
        if not self.running:
            return

        tr = self.get_transaction(session)
        if tr is None:
            # FIXME: return or create a new one?
            return

        tr.begin(session)

    def commit(self, session: Session):
        if not self.running:
            return

        tr = self.get_transaction(session)
        if tr is None:
            return

        tr.commit(session)

    def flush(self, session: Session, flush_context: UOWTransaction):
        # when sqlalchemy is flushing it is done in a sub-transaction,
        # not the root one. So when calling our 'commit' from here
        # we are not in our root transaction, so changes will not be
        # written to repository.
        self.commit(session)

    def rollback(self, session: Session):
        if not self.running:
            return

        tr = self.get_transaction(session)
        if tr is None:
            return

        tr.rollback(session)


class SessionBlobStoreService(Service):
    """A blob store service that is session aware, i.e content is actually
    written or delete at commit time.

    All content is stored using the main :class:`BlobStoreService`.
    """

    name = "session_blob_store"
    AppStateClass = SessionBlobStoreState

    def __init__(self, *args, **kwargs):
        self.__listening = False
        super().__init__(*args, **kwargs)

    def init_app(self, app: Application):
        super().init_app(app)

        path = Path(app.instance_path, "tmp", "files_transactions")
        if not path.exists():
            path.mkdir(0o775, parents=True)

        with app.app_context():
            self.app_state.path = path.resolve()

        if not self.__listening:
            self.start_listening()

    def start_listening(self):
        self.__listening = True
        listen = sa.event.listen
        listen(Session, "after_transaction_create", self.create_transaction)
        listen(Session, "after_transaction_end", self.end_transaction)
        listen(Session, "after_begin", self.begin)
        listen(Session, "after_commit", self.commit)
        listen(Session, "after_flush", self.flush)
        listen(Session, "after_rollback", self.rollback)
        # appcontext_tearing_down.connect(self.clear_transaction, app)

    def _session_for(self, model_or_session: Model | Session) -> Session:
        """Return session instance for object parameter.

        - If parameter is a session instance, it is returned as is.

        - If parameter is a registered model instance, its session will be used.

        - If parameter is a detached model instance, or None, application scoped
          session will be used (db.session()).

        - If parameter is a scoped_session instance, a new session will be
          instaniated.
        """
        if model_or_session is None:
            return db.session

        if isinstance(model_or_session, Session):
            return model_or_session

        if isinstance(model_or_session, sa.orm.scoped_session):
            return model_or_session()

        session = sa.orm.object_session(model_or_session)

        if session is None:
            return db.session

        return session

    # Blob store interface
    def get(
        self, session: Session | Blob, uuid: UUID, default: Path | None = None
    ) -> Path | None:
        # assert isinstance(session, Session)
        _assert_uuid(uuid)

        session = self._session_for(session)
        transaction = self.app_state.get_transaction(session)
        try:
            val = transaction.get(uuid)
        except KeyError:
            return default

        if val is _NULL_MARK:
            val = blob_store.get(uuid, default)

        return val

    def set(
        self,
        session: Session | Blob,
        uuid: UUID,
        content: IO | bytes | str,
        encoding: str = "utf-8",
    ):
        _assert_uuid(uuid)

        session = self._session_for(session)
        transaction = self.app_state.get_transaction(session)
        transaction.set(uuid, content, encoding)

    def delete(self, session: Session | Blob, uuid: UUID):
        _assert_uuid(uuid)

        session = self._session_for(session)
        transaction = self.app_state.get_transaction(session)
        if self.get(session, uuid) is not None:
            transaction.delete(uuid)

    # session event handlers
    @Service.if_running
    def create_transaction(
        self, session: Session, transaction: SessionTransaction
    ) -> Any | None:
        return self.app_state.create_transaction(session, transaction)

    @Service.if_running
    def end_transaction(
        self, session: Session, transaction: SessionTransaction
    ) -> Any | None:
        return self.app_state.end_transaction(session, transaction)

    @Service.if_running
    def begin(
        self, session: Session, transaction: SessionTransaction, connection: Connection
    ) -> Any | None:
        return self.app_state.begin(session)

    @Service.if_running
    def commit(self, session: Session):
        return self.app_state.commit(session)

    @Service.if_running
    def flush(self, session: Session, flush_context: UOWTransaction):
        return self.app_state.flush(session, flush_context)

    @Service.if_running
    def rollback(self, session: Session):
        return self.app_state.rollback(session)


session_blob_store = SessionBlobStoreService()


class BlobStoreTransaction:
    def __init__(self, root_path: Path, parent: BlobStoreTransaction | None = None):
        self.path = root_path / str(uuid1())
        # if parent is not None and parent.cleared:
        #   parent = None

        self._parent = parent
        self._deleted: set[UUID] = set()
        self._set: set[UUID] = set()
        self.__cleared = False

    @property
    def cleared(self) -> bool:
        return self.__cleared

    def __del__(self):
        if not self.cleared:
            self._clear()

    def _clear(self):
        if self.__cleared:
            return

        # make sure transaction is not usable anymore
        if self.path.exists():
            shutil.rmtree(str(self.path))

        del self.path
        del self._deleted
        del self._set
        self.__cleared = True

    def begin(self, session: Session | None = None):
        if not self.path.exists():
            self.path.mkdir(0o700)

    def rollback(self, session: Session | None = None):
        self._clear()

    def commit(self, session: Session | None = None):
        """Merge modified objects into parent transaction.

        Once commited a transaction object is not usable anymore

        :param:session: current sqlalchemy Session
        """
        if self.__cleared:
            return

        if self._parent:
            # nested transaction
            self._commit_parent()
        else:
            self._commit_blob_store()
        self._clear()

    def _commit_blob_store(self):
        assert self._parent is None

        for uuid in self._deleted:
            with contextlib.suppress(KeyError):
                blob_store.delete(uuid)

        for uuid in self._set:
            content = self.path / str(uuid)
            blob_store.set(uuid, content.open("rb"))

    def _commit_parent(self):
        p = self._parent
        assert p
        p._deleted |= self._deleted
        p._deleted -= self._set

        p._set |= self._set
        p._set -= self._deleted

        if self._set:
            p.begin()  # ensure p.path exists

        for uuid in self._set:
            content_path = self.path / str(uuid)
            # content_path.replace is not available with python < 3.3.
            content_path.rename(p.path / str(uuid))

    def _add_to(self, uuid: UUID, dest: set[UUID], other: set[UUID]):
        """Add `item` to `dest` set, ensuring `item` is not present in `other`
        set."""
        _assert_uuid(uuid)
        with contextlib.suppress(KeyError):
            other.remove(uuid)
        dest.add(uuid)

    def delete(self, uuid: UUID):
        self._add_to(uuid, self._deleted, self._set)

    def set(
        self, uuid: UUID, content: IO | bytes | str, encoding: str | None = "utf-8"
    ):
        self.begin()
        self._add_to(uuid, self._set, self._deleted)

        if hasattr(content, "read"):
            content = content.read()

        if isinstance(content, bytes):
            mode = "wb"
            encoding = None
        else:
            mode = "wt"

        dest = self.path / str(uuid)
        with dest.open(mode, encoding=encoding) as f:
            f.write(content)

    def get(self, uuid: UUID) -> Any:
        if uuid in self._deleted:
            raise KeyError(uuid)

        if uuid in self._set:
            path = self.path / str(uuid)
            assert path.exists()
            return path

        if self._parent:
            return self._parent.get(uuid)

        return _NULL_MARK



from __future__ import annotations

from collections.abc import Callable
from datetime import datetime
from itertools import chain

import pytz
import sqlalchemy as sa
import sqlalchemy.orm
from flask import (
    get_template_attribute,
    render_template,
    render_template_string,
    request,
)
from flask_babel import format_date, get_locale
from markupsafe import Markup
from werkzeug.exceptions import InternalServerError
from werkzeug.routing import BuildError

from abilian.core.entities import Entity, all_entity_classes
from abilian.core.extensions import db
from abilian.core.models.subjects import User
from abilian.core.util import local_dt
from abilian.i18n import _
from abilian.services.audit import AuditEntry
from abilian.services.security import SecurityAudit
from abilian.web.admin import AdminPanel
from abilian.web.util import url_for
from abilian.web.views.base import JSONView


def format_date_for_input(date):
    date_fmt = get_locale().date_formats["short"].pattern
    # force numerical months
    date_fmt = date_fmt.replace("MMMM", "MM").replace("MMM", "MM")
    return format_date(date, date_fmt)


class JSONUserSearch(JSONView):
    """Search users by fullname."""

    def data(self, *args, **kwargs) -> dict:
        q: str = kwargs["q"]
        q = q.replace("%", " ").strip().lower()

        if not q or len(q) < 2:
            raise InternalServerError

        query = User.query
        lower = sa.sql.func.lower
        filters = []
        for part in q.split(" "):
            filters.append(
                sa.sql.or_(
                    lower(User.first_name).like(f"{part}%"),
                    lower(User.last_name).like(f"{part}%"),
                )
            )

        filters = sa.sql.and_(*filters) if len(filters) > 1 else filters[0]

        if "@" in q:
            # FIXME: where does this 'part' variable come from ?
            filters = sa.sql.or_(lower(User.email).like(f"%{part}%"), filters)

        query = query.filter(filters).order_by(User.last_name, User.first_name)

        values = query.values(User.id, User.first_name, User.last_name, User.email)
        results = [
            {"id": obj.id, "text": f"{obj.first_name} {obj.last_name} ({obj.email})"}
            for obj in values
        ]

        return {"results": results}


class AuditPanel(AdminPanel):
    """Global audit log.

    Can receive one out of two possible GET parameter, datetime type:

    * before: events that happened immediately before 't' (less recent than 't')
    * after: events that happened immediately after 't' (more recent than 't')
    """

    id = "audit"
    label = "Audit trail"
    icon = "list-alt"

    def install_additional_rules(self, add_url_rule: Callable):
        add_url_rule("/search_users", view_func=JSONUserSearch.as_view("search_users"))

    def get(self) -> str:
        LIMIT = 30

        base_audit_q = AuditEntry.query
        base_security_q = SecurityAudit.query

        before = request.args.get("before")
        after = request.args.get("after")

        # filter on user
        filter_user = None
        user_id = request.args.get("user")
        if user_id:
            user_id = int(user_id)
            filter_user = User.query.get(user_id)
            base_audit_q = base_audit_q.filter(AuditEntry.user == filter_user)
            base_security_q = base_security_q.filter(
                SecurityAudit.manager == filter_user
            )

        # filter by types
        all_classes = sorted(all_entity_classes(), key=lambda c: c.__name__)
        all_types = {e.entity_type for e in all_classes}
        filter_types = set(request.args.getlist("types")) & all_types
        if filter_types:
            if len(filter_types) == 1:
                t = next(iter(filter_types))
                audit_expr = AuditEntry.entity_type == t
                sec_expr = sa.sql.or_(
                    SecurityAudit.object == None, Entity._entity_type == t
                )
            else:
                audit_expr = AuditEntry.entity_type.in_(filter_types)
                sec_expr = sa.sql.or_(
                    SecurityAudit.object == None, Entity._entity_type.in_(filter_types)
                )

            base_audit_q = base_audit_q.filter(audit_expr)
            base_security_q = (
                base_security_q.join(SecurityAudit.object)
                .filter(sec_expr)
                .reset_joinpoint()
            )

        def after_query(q, model, date):
            return q.filter(model.happened_at > date).order_by(model.happened_at.asc())

        def before_query(q, model, date):
            return q.filter(model.happened_at < date).order_by(model.happened_at.desc())

        if after:
            after = datetime.strptime(after, "%Y-%m-%dT%H:%M:%S.%f")
            audit_q = after_query(base_audit_q, AuditEntry, after)
            security_q = after_query(base_security_q, SecurityAudit, after)
        else:
            before = (
                datetime.strptime(before, "%Y-%m-%dT%H:%M:%S.%f")
                if before
                else datetime.utcnow()
            )
            audit_q = before_query(base_audit_q, AuditEntry, before)
            security_q = before_query(base_security_q, SecurityAudit, before)

        audit_entries = (
            audit_q.options(sa.orm.joinedload(AuditEntry.entity)).limit(LIMIT).all()
        )
        security_entries = (
            security_q.options(sa.orm.joinedload(SecurityAudit.object))
            .limit(LIMIT)
            .all()
        )
        # audit_entries = []
        all_entries = sorted(
            chain(
                #
                (AuditEntryPresenter(e) for e in audit_entries),
                #
                (SecurityEntryPresenter(e) for e in security_entries),
            )
        )

        if after:
            all_entries = all_entries[:LIMIT]
        else:
            all_entries = all_entries[-LIMIT:]

        all_entries.reverse()  # event are presented from most to least recent

        # group entries by day
        entries = []
        day_entries: list[BaseEntryPresenter] = []
        current_day = None

        for e in all_entries:
            e_date = e.date

            if e_date.date() != current_day:
                current_day = e_date.date()
                day_entries = []
                entries.append((e_date.date(), day_entries))
            day_entries.append(e)

        top_date = ""
        lowest_date = ""

        if entries:
            # top_date and lowest_date are converted to naive datetime (from UTC), so
            # that isoformat does not include TZ shift (else we should fix strptime
            # above)
            top_date = entries[0][1][0].date.astimezone(pytz.utc).replace(tzinfo=None)
            lowest_date = (
                entries[-1][1][-1].date.astimezone(pytz.utc).replace(tzinfo=None)
            )

            after_queries = (
                after_query(base_audit_q, AuditEntry, top_date),
                after_query(base_security_q, SecurityAudit, top_date),
            )

            before_queries = (
                before_query(base_audit_q, AuditEntry, lowest_date),
                before_query(base_security_q, SecurityAudit, lowest_date).limit(1),
            )

            if not any(q.limit(1).first() is not None for q in after_queries):
                top_date = ""
            else:
                top_date = top_date.isoformat()

            if not any(q.first() is not None for q in before_queries):
                lowest_date = ""
            else:
                lowest_date = lowest_date.isoformat()

        current_date = None
        if entries:
            current_date = format_date_for_input(entries[0][0])

        # build prev/next urls
        url_params = {}
        if filter_user:
            url_params["user"] = filter_user.id
        if filter_types:
            url_params["types"] = next(iter(filter_types))

        ctx = {
            "entries": entries,
            "filter_user": filter_user,
            "all_classes": [(c.__name__, c.entity_type) for c in all_classes],
            "filter_types": filter_types,
            "url_params": url_params,
            "current_date": current_date,
            "top_date": top_date,
            "lowest_date": lowest_date,
        }
        return render_template("admin/audit.html", **ctx)


#
#  Presenters for audit entries listing
#
class BaseEntryPresenter:
    _USER_FMT = (
        '<a href="{{ url_for("social.user", user_id=user.id) }}">' "{{ user.name }}</a>"
    )
    _GROUP_FMT = (
        '<a href="{{ url_for("social.group_home", group_id=group.id)'
        ' }}">{{ group.name }}</a>'
    )

    def __init__(self, user, date):
        self.user = user
        self.date = local_dt(date)

    def __lt__(self, other):
        return self.date < other.date

    def __repr__(self):
        return "{}({}, {} @ {})".format(
            self.__class__.__name__, repr(self.user), repr(self.date), id(self)
        )

    @staticmethod
    def model(model_name):
        return db.Model._decl_class_registry.get(model_name)

    def render(self) -> str:
        raise NotImplementedError


class AuditEntryPresenter(BaseEntryPresenter):
    def __init__(self, entry: AuditEntry):
        assert isinstance(entry, AuditEntry)
        super().__init__(entry.user, entry.happened_at)
        self.entry = entry
        self.entity_deleted = entry.entity is None

    def render(self) -> str:
        render = render_template_string
        e = self.entry
        user = render(self._USER_FMT, user=e.user)
        entity_html = e.entity_name

        if not self.entity_deleted:
            try:
                entity_url = url_for(e.entity)
            except (BuildError, ValueError):
                pass
            else:
                entity_html = Markup(
                    render(
                        '<a href="{{ url }}">{{ entity.path or entity.name }}</a>',
                        url=entity_url,
                        entity=e.entity,
                    )
                )

        if e.type == 0:
            msg = _('{user} created {entity_type} {entity_id} "{entity}"')
        elif e.related or e.op == 1:
            msg = _('{user} made changes on {entity_type} {entity_id} "{entity}"')
        elif e.op == 2:
            msg = _('{user} has deleted {entity_type}: {entity_id} "{entity}"')
        else:
            raise Exception(f"Bad entry type: {e.type}")

        self.msg = Markup(
            msg.format(
                user=user,
                entity=entity_html,
                entity_type=e.entity_type.rsplit(".", 1)[-1],
                entity_id=e.entity_id,
            )
        )
        tmpl = get_template_attribute("admin/_macros.html", "m_audit_entry")
        return tmpl(self)


class SecurityEntryPresenter(BaseEntryPresenter):
    def __init__(self, entry: SecurityAudit):
        assert isinstance(entry, SecurityAudit)
        super().__init__(entry.manager, entry.happened_at)
        self.entry = entry

    def render(self) -> str:
        render = render_template_string
        e = self.entry

        manager = render(
            '<img class="avatar" '
            'src="{{ user_photo_url(user=e.manager, size=16) }}" alt="" />'
            '<a href="'
            '{{ url_for("social.user", user_id=e.manager.id) }}">'
            "{{ e.manager.name }}</a>",
            e=e,
        )

        if self.entry.user:
            principal = render(self._USER_FMT, user=self.entry.user)
        elif self.entry.group:
            principal = render(self._GROUP_FMT, group=self.entry.group)
        else:
            principal = ""

        entity = ""
        if e.object_id:
            entity_url = None
            entity_name = e.object_name
            if e.object:
                entity_name = getattr(e.object, "path", e.object.name)
                entity_url = url_for(e.object)

            entity = render(
                '{%- if url %}<a href="{{ url }}">{%- endif %}'
                "{{ name }}{%- if url %}</a>{%- endif %}",
                url=entity_url,
                name=entity_name,
            )

            if e.op == e.SET_INHERIT:
                msg = _("{manager} has activated inheritance on {entity}")
            elif e.op == e.UNSET_INHERIT:
                msg = _("{manager} has deactivated inheritance on {entity}")
            elif e.op == e.GRANT:
                msg = _(
                    '{manager} has given role "{role}" to {principal} ' "on {entity}"
                )
            elif e.op == e.REVOKE:
                msg = _(
                    '{manager} has revoked role "{role}" from '
                    "{principal} on {entity}"
                )
            else:
                raise Exception(f"Invalid entity op: {e.op}")
        else:
            if e.op == e.GRANT:
                msg = _('{manager} has given role "{role}" to {principal}')
            elif e.op == e.REVOKE:
                msg = _('{manager} has revoked role "{role}" from {principal}')
            else:
                raise Exception(f"Invalid entity op: {e.op}")

        self.msg = Markup(
            msg.format(manager=manager, principal=principal, role=e.role, entity=entity)
        )
        tmpl = get_template_attribute("admin/_macros.html", "m_security_entry")
        return tmpl(self)



""""""
from __future__ import annotations

import abc
from typing import Any

from sqlalchemy import CheckConstraint, Column, ForeignKey, Integer, UnicodeText
from sqlalchemy.orm import backref, relationship

from abilian.core.entities import Entity, EntityQuery
from abilian.services.security import CREATE, DELETE, WRITE, Anonymous, Owner

#: name of backref on target :class:`Entity` object
ATTRIBUTE = "__comments__"


class Commentable(metaclass=abc.ABCMeta):
    id: int


def register(cls: type[Entity]) -> type:
    """Register an :class:`Entity` as a commentable class.

    Can be used as a class decorator:

    .. code-block:: python

      @comment.register
      class MyContent(Entity):
          ...
    """
    if not issubclass(cls, Entity):
        raise ValueError("Class must be a subclass of abilian.core.entities.Entity")

    Commentable.register(cls)
    return cls


def is_commentable(obj_or_class: Any) -> bool:
    """
    :param obj_or_class: a class or instance
    """
    if isinstance(obj_or_class, type):
        return issubclass(obj_or_class, Commentable)

    if not isinstance(obj_or_class, Commentable):
        return False

    if obj_or_class.id is None:
        return False

    return True


def for_entity(obj, check_commentable=False):
    """Return comments on an entity."""
    if check_commentable and not is_commentable(obj):
        return []

    return getattr(obj, ATTRIBUTE)


class CommentQuery(EntityQuery):
    def all(self):
        return EntityQuery.all(self.order_by(Comment.created_at))


class Comment(Entity):
    """A Comment related to an :class:`Entity`."""

    __default_permissions__ = {WRITE: {Owner}, DELETE: {Owner}, CREATE: {Anonymous}}

    entity_id = Column(Integer, ForeignKey(Entity.id), nullable=False)

    #: Commented entity
    entity = relationship(
        Entity,
        lazy="immediate",
        foreign_keys=[entity_id],
        backref=backref(
            ATTRIBUTE,
            lazy="select",
            order_by="Comment.created_at",
            cascade="all, delete-orphan",
        ),
    )

    #: comment's main content
    body = Column(UnicodeText(), CheckConstraint("trim(body) != ''"), nullable=False)

    query_class = CommentQuery

    @property
    def history(self):
        return self.meta.get("abilian.core.models.comment", {}).get("history", [])

    def __repr__(self):
        class_ = self.__class__
        mod_ = class_.__module__
        classname = class_.__name__
        return "<{}.{} instance at 0x{:x} entity id={!r} date={}".format(
            mod_, classname, id(self), self.entity_id, self.created_at
        )



from __future__ import annotations

from flask_babel import lazy_gettext as _l

from abilian.core.util import BasePresenter


class CommunityPresenter(BasePresenter):
    @property
    def breadcrumbs(self):
        return [
            {"label": _l("Communities"), "path": "/communities/"},
            {"label": self._model.name},
        ]



""""""
from __future__ import annotations

from flask import current_app
from flask_debugtoolbar.panels import DebugPanel

from abilian.core.util import fqcn
from abilian.i18n import _
from abilian.services import get_service
from abilian.web.action import actions


class IndexedTermsDebugPanel(DebugPanel):
    """A panel to display term values found in index for "current" object.

    FIXME: this notion of "current" object should formalized in
    abilian.app.Application
    """

    name = "IndexedTerms"

    @property
    def current_obj(self):
        return actions.context.get("object")

    @property
    def has_content(self):
        obj = self.current_obj
        return (
            obj is not None
            and hasattr(obj, "object_type")
            and hasattr(obj, "id")
            and obj.id is not None
        )

    def nav_title(self):
        return _("Indexed Terms")

    def nav_subtitle(self):
        """Subtitle showing until title in toolbar."""
        obj = self.current_obj
        if not obj:
            return _("No current object")

        try:
            return f"{obj.__class__.__name__}(id={obj.id})"
        except Exception:
            return ""

    def title(self):
        return _("Indexed Terms")

    def url(self):
        return ""

    def content(self):
        obj = self.current_obj

        index_service = get_service("indexing")
        index = index_service.app_state.indexes["default"]
        schema = index.schema
        context = self.context.copy()
        context["schema"] = schema
        context["sorted_fields"] = sorted(schema.names())

        adapter = index_service.adapted.get(fqcn(obj.__class__))
        if adapter and adapter.indexable:
            doc = context["current_document"] = index_service.get_document(obj, adapter)
            indexed = {}
            for name, field in schema.items():
                value = doc.get(name)
                indexed[name] = None
                if value and field.analyzer and field.format:
                    indexed[name] = list(field.process_text(value))
            context["current_indexed"] = indexed
            context["current_keys"] = sorted(set(doc) | set(indexed))

        with index.searcher() as search:
            document = search.document(object_key=obj.object_key)

        sorted_keys = sorted(document) if document is not None else None

        context.update({"document": document, "sorted_keys": sorted_keys})

        jinja_env = current_app.jinja_env
        jinja_env.filters.update(self.jinja_env.filters)
        template = jinja_env.get_or_select_template("debug_panels/indexing_panel.html")
        return template.render(context)



from __future__ import annotations

from typing import cast

from flask_login import current_user, login_user
from pytest import fixture
from sqlalchemy.orm import Session

from abilian.app import Application as BaseApplication
from abilian.core.models.subjects import User
from abilian.core.sqlalchemy import SQLAlchemy
from abilian.services import get_service, security_service
from abilian.services.security import SecurityService

from .models import UserPreference
from .panel import PreferencePanel
from .service import PreferenceService


class VisiblePanel(PreferencePanel):
    id = label = "visible"

    def is_accessible(self) -> bool:
        return True

    def get(self):
        return "Visible"


class AdminPanel(PreferencePanel):
    id = label = "admin"

    def is_accessible(self) -> bool:
        security = cast(SecurityService, get_service("security"))
        return security.has_role(current_user, "admin")

    def get(self):
        return "Admin"


class Application(BaseApplication):
    def init_extensions(self):
        super().init_extensions()
        prefs = cast(PreferenceService, self.services["preferences"])
        prefs.app_state.panels = []
        prefs.register_panel(VisiblePanel(), self)
        prefs.register_panel(AdminPanel(), self)


@fixture()
def app(config: type) -> Application:
    app = Application()
    app.setup(config)
    return app


def test_preferences(app: Application, session: Session):
    user = User(email="test@example.com")
    assert UserPreference.query.all() == []

    preference_service = PreferenceService()

    preferences = preference_service.get_preferences(user)
    assert preferences == {}

    preference_service.set_preferences(user, digest="daily")
    session.flush()

    preferences = preference_service.get_preferences(user)
    assert preferences == {"digest": "daily"}

    preference_service.clear_preferences(user)
    session.flush()

    preferences = preference_service.get_preferences(user)
    assert preferences == {}
    assert UserPreference.query.all() == []


def test_preferences_with_various_types(app: Application, session: Session):
    user = User(email="test@example.com")
    preference_service = PreferenceService()

    preference_service.set_preferences(user, some_int=1)
    session.flush()
    preferences = preference_service.get_preferences(user)
    assert preferences == {"some_int": 1}

    preference_service.set_preferences(user, some_bool=True)
    session.flush()
    preferences = preference_service.get_preferences(user)
    assert preferences == {"some_int": 1, "some_bool": True}


def test_visible_panels(app: Application, db: SQLAlchemy):
    user = User(email="test@example.com")

    with app.test_request_context():
        security_service.start()
        login_user(user)

        for cp in app.template_context_processors["preferences"]:
            ctx = cp()
            if "menu" in ctx:
                break

        expected = ["preferences.visible"]
        assert [p["endpoint"] for p in ctx["menu"]] == expected

        security_service.grant_role(user, "admin")
        ctx = cp()
        expected = ["preferences.visible", "preferences.admin"]
        assert [p["endpoint"] for p in ctx["menu"]] == expected



from __future__ import annotations

import uuid
from collections.abc import Callable
from pathlib import Path
from typing import TYPE_CHECKING

import pkg_resources
from loguru import logger

from abilian.core.util import fqcn

if TYPE_CHECKING:
    from abilian.sbe.app import Application


STATIC_DIR = pkg_resources.resource_filename(__name__, "static")
LESSCSS_FILE = str(Path(STATIC_DIR, "less", "abilian-sbe.less"))
JS = (
    "js/sbe-datatable.js",
    "js/folder.js",
    "js/folder_edit.js",
    "js/folder_gallery.js",
    "js/folder_upload.js",
    "js/document_viewer.js",
    "vendor/bootstrap-tagsinput.js",
    "vendor/jquery.fileapi.js",
)


class AbilianSBE:
    """Base extension required by abilian.sbe.apps."""

    def __init__(self, app: Application | None = None) -> None:
        if app is not None:
            self.init_app(app)

    def init_app(self, app: Application) -> None:
        # False: it's ok if antivirus task was run but service couldn't get a
        # result
        app.config.setdefault("ANTIVIRUS_CHECK_REQUIRED", False)
        app.config.setdefault("SBE_FORUM_REPLY_BY_MAIL", False)

        if FQCN in app.extensions:
            return

        app.extensions[FQCN] = self

        # register i18n
        app.extensions["babel"].add_translations("abilian.sbe")

        # sbe static assets
        app.add_static_url("abilian/sbe", STATIC_DIR, endpoint="abilian_sbe_static")
        app.extensions["webassets"].append_path(
            STATIC_DIR, f"{app.static_url_path}/abilian/sbe"
        )

        app.register_asset("js", *JS)
        app.register_asset("css", LESSCSS_FILE)

        # Jinja
        logger.info("Register jinja context processors")
        app.context_processor(inject_template_utils)


FQCN = fqcn(AbilianSBE)
sbe = AbilianSBE()


def inject_template_utils() -> dict[str, Callable]:
    return {"uuid": uuid.uuid1}



from __future__ import annotations

from flask import Flask

from abilian.core.models.subjects import Group, User
from abilian.core.sqlalchemy import SQLAlchemy


def test_non_ascii_password():
    """Ensure we can store and test non-ascii password without any
    UnicodeEncodeError."""
    user = User()

    user.set_password("Hé")

    if not isinstance(user.password, str):
        # when actually retrieved from database, it should be Unicode
        user.password = str(user.password)

    assert user.authenticate("Hé")


def test_user(app: Flask, db: SQLAlchemy):
    user = User(email="test@test.com")
    db.session.add(user)
    db.session.flush()

    assert not user.is_online


def test_group(app: Flask, db: SQLAlchemy):
    group = Group(name="test_group")
    db.session.add(group)
    db.session.flush()

    assert group.members_count == 0


def test_follow(app: Flask, db: SQLAlchemy):
    user1 = User(email="test1@test.com")
    user2 = User(email="test2@test.com")
    db.session.add(user1)
    db.session.add(user2)
    db.session.flush()

    assert not user1.is_following(user2)

    user1.follow(user2)
    assert user1.is_following(user2)

    user1.unfollow(user2)
    assert not user1.is_following(user2)


def test_group_membership(app: Flask, db: SQLAlchemy):
    user = User(email="test@test.com")
    db.session.add(user)
    group = Group(name="test_group")
    db.session.add(group)
    db.session.flush()

    assert not user.is_member_of(group)

    user.join(group)
    assert user.is_member_of(group)
    assert user in group.members

    user.leave(group)
    assert not user.is_member_of(group)



from __future__ import annotations

from typing import Any

from sqlalchemy.engine.interfaces import Dialect
from sqlalchemy.types import String, TypeDecorator


class ValueSingletonMeta(type):
    __instances__: dict[str, Any]

    def __new__(
        mcs: type[ValueSingletonMeta],
        name: str,
        bases: tuple[type],
        dct: dict[str, Any],
    ) -> ValueSingletonMeta:
        dct["__instances__"] = {}
        dct.setdefault("__slots__", ())
        new_type = type.__new__(mcs, name, bases, dct)
        return new_type

    def __call__(cls, value: Any, *args, **kwargs) -> Any:
        if isinstance(value, cls):
            return value

        if value not in cls.__instances__:
            value_instance = super().__call__(value, *args, **kwargs)
            cls.__instances__[getattr(value_instance, cls.attr)] = value_instance

        result = cls.__instances__[value.lower()]

        # from devtools import debug
        # debug(result)
        # assert False
        return result


class UniqueName(metaclass=ValueSingletonMeta):
    """Base class to create singletons from strings.

    A subclass of :class:`UniqueName` defines a namespace.
    """

    __slots__ = ("_hash", "__name")
    attr = "name"

    def __init__(self, name: str):
        self.__name = str(name).strip().lower()
        self._hash = hash(self.__name)

    @property
    def name(self) -> str:
        return self.__name

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name!r})"

    def __str__(self) -> str:
        return self.name

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self._hash == other._hash
        return self.__name == str(other)

    def __hash__(self) -> int:
        return self._hash


class UniqueNameType(TypeDecorator):
    """Sqlalchemy type that stores a subclass of :class:`UniqueName`

    Usage::
    class MySingleton(UniqueName):
        pass

    class MySingletonType(UniqueNameType):
        Type = MySingleton
    """

    impl = String
    Type: type
    default_max_length = 100

    def __init__(self, *args, **kwargs):
        assert self.Type is not None
        kwargs.setdefault("length", self.default_max_length)
        super().__init__(*args, **kwargs)

    def process_bind_param(self, value: Any, dialect: Dialect) -> str | None:
        if value is not None:
            value = str(value)
        return value

    def process_result_value(self, value: str | None, dialect: Dialect) -> Any:
        if value is not None:
            value = self.Type(value)
        return value



from __future__ import annotations

import pytest
from flask.ctx import AppContext
from sqlalchemy.orm import Session

from abilian.sbe.app import Application, create_app
from abilian.sbe.apps.documents.models import Document, Folder, icon_for


@pytest.fixture()
def app(config: type) -> Application:
    return create_app(config=config)


def check_editable(object: Document | Folder) -> None:
    if hasattr(object, "__editable__"):
        for k in object.__editable__:
            assert hasattr(object, k)


def test_title_prevails() -> None:
    f = Folder(name="name", title="title")
    assert f.title == "title"
    assert f.name == "title"

    f = Folder(name="name", title=None)
    assert f.title == "name"
    assert f.name == "name"

    f = Folder(name="name")
    assert f.title == "name"
    assert f.name == "name"


def test_folder_editables() -> None:
    root = Folder(title="/")
    check_editable(root)


def test_folder_can_create_documents() -> None:
    root = Folder(title="/")

    document = root.create_document("doc")

    assert len(root.children) == 1
    assert document.title == "doc"
    assert document.path == "/doc"
    assert document in root.documents
    assert document in root.children
    assert document.parent == root


def test_folder_can_create_subfolders() -> None:
    root = Folder(title="/")

    subfolder = root.create_subfolder("folder")
    assert len(root.children) == 1
    assert subfolder.title == "folder"
    assert subfolder.path == "/folder"
    assert subfolder in root.subfolders
    assert subfolder in root.children
    assert subfolder.parent == root


def test_nested_subobjects() -> None:
    root = Folder(title="/")
    subfolder = root.create_subfolder("folder1")
    subsubfolder = subfolder.create_subfolder("folder2")
    document = subfolder.create_document("doc")

    assert len(root.children) == 1
    assert len(subfolder.children) == 2
    assert subsubfolder.title == "folder2"
    assert subsubfolder.path == "/folder1/folder2"
    assert document.title == "doc"
    assert document.path == "/folder1/doc"

    assert root.get_object_by_path("/folder1") == subfolder
    assert root.get_object_by_path("/folder1/folder2") == subsubfolder
    assert root.get_object_by_path("/folder1/doc") == document


def test_folder_is_clonable() -> None:
    root = Folder(title="/")
    clone = root.clone()

    assert clone.title == root.title
    assert clone.path == root.path


def test_document_editables() -> None:
    doc = Document()
    check_editable(doc)


def test_content_length(session: Session) -> None:
    doc = Document(title="toto")
    doc.set_content(b"tototiti", "application/binary")
    assert doc.content_length == len("tototiti")


def test_document_is_clonable(session: Session) -> None:
    root = Folder(title="/")
    doc = root.create_document(title="toto")
    doc.content = b"tototiti"

    clone = doc.clone()
    assert clone.title == doc.title
    assert clone.content == doc.content


def test_document_has_an_icon(app_context: AppContext) -> None:
    root = Folder(title="/")
    doc = root.create_document(title="toto")
    doc.content_type = "image/jpeg"
    filename = doc.icon.split("/")[-1]
    assert filename in ("jpg.png", "jpeg.png"), doc.icon


def test_icon_from_mime_type(app_context: AppContext) -> None:
    icon = icon_for("text/html")
    filename = icon.split("/")[-1]
    assert filename in ("html.png", "htm.png"), icon



""""""
from __future__ import annotations

from collections import OrderedDict
from functools import partial

import whoosh
import whoosh.highlight
import whoosh.searching
import whoosh.sorting
from flask import Blueprint, current_app, g, render_template, request, url_for

from abilian.i18n import _
from abilian.services import get_service
from abilian.web import views
from abilian.web.action import Endpoint
from abilian.web.nav import BreadcrumbItem

BOOTSTRAP_MARKUP_HIGHLIGHTER = whoosh.highlight.HtmlFormatter(
    tagname="mark", classname="", termclass="term-", between="[…]"
)

RESULTS_FRAGMENTER = whoosh.highlight.SentenceFragmenter()

#
# Some hardcoded settings (for now)
#
PAGE_SIZE = 20
MAX_LIVE_RESULTS_PER_CLASS = 5

search = Blueprint(
    "search", __name__, url_prefix="/search", template_folder="templates"
)
route = search.route


def friendly_fqcn(fqcn):
    return fqcn.rsplit(".", 1)[-1]


@search.url_value_preprocessor
def init_search(endpoint, values):
    q = request.args.get("q", "")
    try:
        page = max(int(request.args.get("page", 1)), 1)
    except Exception:
        page = 1

    g.breadcrumb.append(
        BreadcrumbItem(
            label=f'"{q}"', icon="search", url=Endpoint("search.search_main", q=q)
        )
    )

    page_kw = OrderedDict(q=q)
    object_types = request.args.getlist("object_type")

    if object_types:
        page_kw["object_type"] = object_types
        g.breadcrumb.append(
            BreadcrumbItem(
                label=" | ".join(friendly_fqcn(name) for name in object_types),
                url=Endpoint("search.search_main", **page_kw),
            )
        )

    if page > 1:
        g.breadcrumb.append(
            BreadcrumbItem(
                label=str(page),
                url=Endpoint("search.search_main", page=page, **page_kw),
            )
        )

    values["q"] = q
    values["page"] = page


@search.context_processor
def install_hit_to_url():
    return {"url_for_hit": current_app.extensions["indexing"].url_for_hit}


_COUNT_OBJECT_TYPE_FACET = whoosh.sorting.FieldFacet(
    "object_type", maptype=whoosh.sorting.Count
)


@route("")
def search_main(q="", page=1):
    svc = get_service("indexing")
    q = q.strip()
    page = int(request.args.get("page", page))
    search_kwargs = {"limit": page * PAGE_SIZE}
    page_url_kw = OrderedDict(q=q)

    search_kwargs["groupedby"] = _COUNT_OBJECT_TYPE_FACET
    results = svc.search(q, **search_kwargs)

    filtered_by_type = sorted(request.args.getlist("object_type"))
    if filtered_by_type:
        page_url_kw["object_type"] = filtered_by_type

    page_url = partial(url_for, ".search_main", **page_url_kw)

    # get facets groups
    by_object_type = []
    try:
        facet_group = results.groups("object_type")
    except KeyError:
        pass
    else:
        for typename, count in facet_group.items():
            is_active = typename in filtered_by_type
            classname = friendly_fqcn(typename)
            link = page_url(object_type=typename)
            by_object_type.append((classname, count, link, is_active))

        by_object_type.sort(key=lambda t: t[0])

    if by_object_type:
        # Insert 'all' to clear all filters
        is_active = len(filtered_by_type) == 0
        all_types = (_("All"), len(results), page_url(object_type=()), is_active)
        by_object_type.insert(0, all_types)

    if filtered_by_type:
        # FIXME: sanitize input
        results = svc.search(q, object_types=filtered_by_type, **search_kwargs)

    results.formatter = BOOTSTRAP_MARKUP_HIGHLIGHTER
    results.fragmenter = RESULTS_FRAGMENTER

    # paginate results
    results_count = len(results)

    # results.pagecount must be ignored when query is filtered: it ignores
    # filtered_count
    pagecount = 1 + results_count // PAGE_SIZE
    page = min(page, pagecount)
    results = whoosh.searching.ResultsPage(results, page, PAGE_SIZE)
    page = results.pagenum
    first_page = page_url(page=1)
    last_page = page_url(page=pagecount)
    prev_page = page_url(page=page - 1) if page > 1 else None
    next_page = page_url(page=page + 1) if page < pagecount else None

    page_min = max(page - 2, 1)
    page_max = min(page + 4, pagecount)
    next_pages_numbered = [
        (index, page_url(page=index)) for index in range(page_min, page_max + 1)
    ]

    ctx = {
        "q": q,
        "results": results,
        "results_count": results_count,
        "pagecount": pagecount,
        "filtered_by_type": filtered_by_type,
        "by_object_type": by_object_type,
        "prev_page": prev_page,
        "next_page": next_page,
        "first_page": first_page,
        "last_page": last_page,
        "next_pages_numbered": next_pages_numbered,
        "friendly_fqcn": friendly_fqcn,
    }
    return render_template("search/search.html", **ctx)


class Live(views.JSONView):
    """JSON response for live search."""

    def data(self, q="", page=None, *args, **kwargs):
        if not q:
            q = ""
        svc = get_service("indexing")
        url_for_hit = svc.app_state.url_for_hit
        search_kwargs = {"facet_by_type": 5}
        results = svc.search(q, **search_kwargs)
        datasets = {}

        for typename, docs in results.items():
            dataset = []
            for doc in docs:
                # Because it happens sometimes
                try:
                    name = doc["name"]
                except KeyError:
                    continue

                d = {"name": name}
                url = url_for_hit(doc, None)
                if url is not None:
                    d["url"] = url
                    dataset.append(d)
            datasets[typename] = dataset

        return {"results": datasets}


route("/live")(Live.as_view("live"))
